# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import torch
from torch.distributions import constraints

from pyro.distributions.torch import Categorical
from pyro.distributions.torch_distribution import TorchDistribution
from pyro.distributions.util import copy_docs_from


@copy_docs_from(TorchDistribution)
class Empirical(TorchDistribution):
    r"""
    Empirical distribution associated with the sampled data. Note that the shape
    requirement for `log_weights` is that its shape must match the leftmost shape
    of `samples`. Samples are aggregated along the ``aggregation_dim``, which is
    the rightmost dim of `log_weights`.

    Example:

    >>> emp_dist = Empirical(torch.randn(2, 3, 10), torch.ones(2, 3))
    >>> emp_dist.batch_shape
    torch.Size([2])
    >>> emp_dist.event_shape
    torch.Size([10])

    >>> single_sample = emp_dist.sample()
    >>> single_sample.shape
    torch.Size([2, 10])
    >>> batch_sample = emp_dist.sample((100,))
    >>> batch_sample.shape
    torch.Size([100, 2, 10])

    >>> emp_dist.log_prob(single_sample).shape
    torch.Size([2])
    >>> # Vectorized samples cannot be scored by log_prob.
    >>> with pyro.validation_enabled():
    ...     emp_dist.log_prob(batch_sample).shape
    Traceback (most recent call last):
    ...
    ValueError: ``value.shape`` must be torch.Size([2, 10])

    :param torch.Tensor samples: samples from the empirical distribution.
    :param torch.Tensor log_weights: log weights (optional) corresponding
        to the samples.
    """

    arg_constraints = {}
    support = constraints.real
    has_enumerate_support = True

    def __init__(self, samples, log_weights, validate_args=None):
        self._samples = samples
        self._log_weights = log_weights
        sample_shape, weight_shape = samples.size(), log_weights.size()
        if (
            weight_shape > sample_shape
            or weight_shape != sample_shape[: len(weight_shape)]
        ):
            raise ValueError(
                "The shape of ``log_weights`` ({}) must match "
                "the leftmost shape of ``samples`` ({})".format(
                    weight_shape, sample_shape
                )
            )
        self._aggregation_dim = log_weights.dim() - 1
        event_shape = sample_shape[len(weight_shape) :]
        self._categorical = Categorical(logits=self._log_weights)
        super().__init__(
            batch_shape=weight_shape[:-1],
            event_shape=event_shape,
            validate_args=validate_args,
        )

    @property
    def sample_size(self):
        """
        Number of samples that constitute the empirical distribution.

        :return int: number of samples collected.
        """
        return self._log_weights.numel()

    def sample(self, sample_shape=torch.Size()):
        sample_idx = self._categorical.sample(
            sample_shape
        )  # sample_shape x batch_shape
        # reorder samples to bring aggregation_dim to the front:
        # batch_shape x num_samples x event_shape -> num_samples x batch_shape x event_shape
        samples = (
            self._samples.unsqueeze(0)
            .transpose(0, self._aggregation_dim + 1)
            .squeeze(self._aggregation_dim + 1)
        )
        # make sample_idx.shape compatible with samples.shape: sample_shape_numel x batch_shape x event_shape
        sample_idx = sample_idx.reshape(
            (-1,) + self.batch_shape + (1,) * len(self.event_shape)
        )
        sample_idx = sample_idx.expand((-1,) + samples.shape[1:])
        return samples.gather(0, sample_idx).reshape(sample_shape + samples.shape[1:])

    def log_prob(self, value):
        """
        Returns the log of the probability mass function evaluated at ``value``.
        Note that this currently only supports scoring values with empty
        ``sample_shape``.

        :param torch.Tensor value: scalar or tensor value to be scored.
        """
        if self._validate_args:
            if value.shape != self.batch_shape + self.event_shape:
                raise ValueError(
                    "``value.shape`` must be {}".format(
                        self.batch_shape + self.event_shape
                    )
                )
        if self.batch_shape:
            value = value.unsqueeze(self._aggregation_dim)
        selection_mask = self._samples.eq(value)
        # Get a mask for all entries in the ``weights`` tensor
        # that correspond to ``value``.
        for _ in range(len(self.event_shape)):
            selection_mask = selection_mask.min(dim=-1)[0]
        selection_mask = selection_mask.type(self._categorical.probs.type())
        return (self._categorical.probs * selection_mask).sum(dim=-1).log()

    def _weighted_mean(self, value, keepdim=False):
        weights = self._log_weights.reshape(
            self._log_weights.size()
            + torch.Size([1] * (value.dim() - self._log_weights.dim()))
        )
        dim = self._aggregation_dim
        max_weight = weights.max(dim=dim, keepdim=True)[0]
        relative_probs = (weights - max_weight).exp()
        return (value * relative_probs).sum(
            dim=dim, keepdim=keepdim
        ) / relative_probs.sum(dim=dim, keepdim=keepdim)

    @property
    def event_shape(self):
        return self._event_shape

    @property
    def mean(self):
        if self._samples.dtype in (torch.int32, torch.int64):
            raise ValueError(
                "Mean for discrete empirical distribution undefined. "
                + "Consider converting samples to ``torch.float32`` "
                + "or ``torch.float64``. If these are samples from a "
                + "`Categorical` distribution, consider converting to a "
                + "`OneHotCategorical` distribution."
            )
        return self._weighted_mean(self._samples)

    @property
    def variance(self):
        if self._samples.dtype in (torch.int32, torch.int64):
            raise ValueError(
                "Variance for discrete empirical distribution undefined. "
                + "Consider converting samples to ``torch.float32`` "
                + "or ``torch.float64``. If these are samples from a "
                + "`Categorical` distribution, consider converting to a "
                + "`OneHotCategorical` distribution."
            )
        mean = self.mean.unsqueeze(self._aggregation_dim)
        deviation_squared = torch.pow(self._samples - mean, 2)
        return self._weighted_mean(deviation_squared)

    @property
    def log_weights(self):
        return self._log_weights

    def enumerate_support(self, expand=True):
        # Empirical does not support batching, so expanding is a no-op.
        return self._samples



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

from torch.distributions import constraints
from torch.distributions.transforms import Transform
from torch.nn.functional import softplus


def softplus_inv(y):
    return y + y.neg().expm1().neg().log()


# Backport of https://github.com/pytorch/pytorch/pull/52300
class SoftplusTransform(Transform):
    r"""
    Transform via the mapping :math:`\text{Softplus}(x) = \log(1 + \exp(x))`.
    """

    domain = constraints.real
    codomain = constraints.positive
    bijective = True
    sign = +1

    def __eq__(self, other):
        return isinstance(other, SoftplusTransform)

    def _call(self, x):
        return softplus(x)

    def _inverse(self, y):
        return softplus_inv(y)

    def log_abs_det_jacobian(self, x, y):
        return -softplus(-x)


class SoftplusLowerCholeskyTransform(Transform):
    """
    Transform from unconstrained matrices to lower-triangular matrices with
    nonnegative diagonal entries. This is useful for parameterizing positive
    definite matrices in terms of their Cholesky factorization.
    """

    domain = constraints.independent(constraints.real, 2)
    codomain = constraints.lower_cholesky

    def __eq__(self, other):
        return isinstance(other, SoftplusLowerCholeskyTransform)

    def _call(self, x):
        diag = softplus(x.diagonal(dim1=-2, dim2=-1))
        return x.tril(-1) + diag.diag_embed()

    def _inverse(self, y):
        diag = softplus_inv(y.diagonal(dim1=-2, dim2=-1))
        return y.tril(-1) + diag.diag_embed()


__all__ = [
    "SoftplusTransform",
    "SoftplusLowerCholeskyTransform",
]



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import warnings
from typing import Callable, Dict

from torch.autograd import grad

# Registry for exception handlers that can be used to catch certain failures
# during computation of `potential_fn` within `potential_grad`.
_EXCEPTION_HANDLERS: Dict[str, Callable[[Exception], bool]] = {}


def velocity_verlet(
    z, r, potential_fn, kinetic_grad, step_size, num_steps=1, z_grads=None
):
    r"""
    Second order symplectic integrator that uses the velocity verlet algorithm.

    :param dict z: dictionary of sample site names and their current values
        (type :class:`~torch.Tensor`).
    :param dict r: dictionary of sample site names and corresponding momenta
        (type :class:`~torch.Tensor`).
    :param callable potential_fn: function that returns potential energy given z
        for each sample site. The negative gradient of the function with respect
        to ``z`` determines the rate of change of the corresponding sites'
        momenta ``r``.
    :param callable kinetic_grad: a function calculating gradient of kinetic energy
        w.r.t. momentum variable.
    :param float step_size: step size for each time step iteration.
    :param int num_steps: number of discrete time steps over which to integrate.
    :param torch.Tensor z_grads: optional gradients of potential energy at current ``z``.
    :return tuple (z_next, r_next, z_grads, potential_energy): next position and momenta,
        together with the potential energy and its gradient w.r.t. ``z_next``.
    """
    z_next = z.copy()
    r_next = r.copy()
    for _ in range(num_steps):
        z_next, r_next, z_grads, potential_energy = _single_step_verlet(
            z_next, r_next, potential_fn, kinetic_grad, step_size, z_grads
        )
    return z_next, r_next, z_grads, potential_energy


def _single_step_verlet(z, r, potential_fn, kinetic_grad, step_size, z_grads=None):
    r"""
    Single step velocity verlet that modifies the `z`, `r` dicts in place.
    """

    z_grads = potential_grad(potential_fn, z)[0] if z_grads is None else z_grads

    for site_name in r:
        r[site_name] = r[site_name] + 0.5 * step_size * (
            -z_grads[site_name]
        )  # r(n+1/2)

    r_grads = kinetic_grad(r)
    for site_name in z:
        z[site_name] = z[site_name] + step_size * r_grads[site_name]  # z(n+1)

    z_grads, potential_energy = potential_grad(potential_fn, z)
    for site_name in r:
        r[site_name] = r[site_name] + 0.5 * step_size * (-z_grads[site_name])  # r(n+1)

    return z, r, z_grads, potential_energy


def potential_grad(potential_fn, z):
    """
    Gradient of `potential_fn` w.r.t. parameters z.

    :param potential_fn: python callable that takes in a dictionary of parameters
        and returns the potential energy.
    :param dict z: dictionary of parameter values keyed by site name.
    :return: tuple of `(z_grads, potential_energy)`, where `z_grads` is a dictionary
        with the same keys as `z` containing gradients and potential_energy is a
        torch scalar.
    """
    z_keys, z_nodes = zip(*z.items())
    for node in z_nodes:
        node.requires_grad_(True)
    try:
        potential_energy = potential_fn(z)
    # handle exceptions as defined in the exception registry
    except Exception as e:
        if any(h(e) for h in _EXCEPTION_HANDLERS.values()):
            grads = {k: v.new_zeros(v.shape) for k, v in z.items()}
            return grads, z_nodes[0].new_tensor(float("nan"))
        else:
            raise e
    grads = grad(potential_energy, z_nodes)
    for node in z_nodes:
        node.requires_grad_(False)
    return dict(zip(z_keys, grads)), potential_energy.detach()


def register_exception_handler(
    name: str, handler: Callable[[Exception], bool], warn_on_overwrite: bool = True
) -> None:
    """
    Register an exception handler for handling (primarily numerical) errors
    when evaluating the potential function.

    :param name: name of the handler (must be unique).
    :param handler: A callable mapping an Exception to a boolean. Exceptions
        that evaluate to true in any of the handlers are handled in the computation
        of the potential energy.
    :param warn_on_overwrite: If True, warns when overwriting a handler already
        registered under the provided name.
    """
    if name in _EXCEPTION_HANDLERS and warn_on_overwrite:
        warnings.warn(
            f"Overwriting Exception handler already registered under key {name}.",
            RuntimeWarning,
        )
    _EXCEPTION_HANDLERS[name] = handler


def _handle_torch_singular(exception: Exception) -> bool:
    """Exception handler for errors thrown on (numerically) singular matrices."""
    # the actual type of the exception thrown is torch._C._LinAlgError
    if isinstance(exception, RuntimeError):
        msg = str(exception)
        return "singular" in msg or "input is not positive-definite" in msg
    return False


# Register default exception handler
register_exception_handler("torch_singular", _handle_torch_singular)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import os
import urllib

import pandas as pd

from pyro.contrib.examples.util import _mkdir_p, get_data_directory

DATA = get_data_directory(__file__)

# https://finance.yahoo.com/quote/%5EGSPC/history/
CACHE_URL = "https://d2hg8soec8ck9v.cloudfront.net/datasets/snp500.csv.bz2"


def load_snp500():
    """
    Loads pandas dataframe of S&P 500 daily values from 1927-12-30 thru 2020-01-10.
    """
    _mkdir_p(DATA)
    filename = os.path.join(DATA, "snp500.csv.bz2")
    if not os.path.exists(filename):
        urllib.request.urlretrieve(CACHE_URL, filename)
    df = pd.read_csv(filename)
    return df



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

"""
This file contains reimplementations of some of Pyro's core enumeration machinery,
which should eventually be drop-in replacements for the current versions.
"""
import functools
import math
from collections import OrderedDict

import funsor
import torch

import pyro.poutine.runtime
import pyro.poutine.util
from pyro.contrib.funsor.handlers.named_messenger import NamedMessenger
from pyro.contrib.funsor.handlers.primitives import to_data, to_funsor
from pyro.contrib.funsor.handlers.replay_messenger import ReplayMessenger
from pyro.contrib.funsor.handlers.trace_messenger import TraceMessenger
from pyro.poutine.escape_messenger import EscapeMessenger
from pyro.poutine.subsample_messenger import _Subsample

funsor.set_backend("torch")


@functools.singledispatch
def _get_support_value(funsor_dist, name, **kwargs):
    raise ValueError(
        "Could not extract point from {} at name {}".format(funsor_dist, name)
    )


@_get_support_value.register(funsor.cnf.Contraction)
def _get_support_value_contraction(funsor_dist, name, **kwargs):
    delta_terms = [
        v
        for v in funsor_dist.terms
        if isinstance(v, funsor.delta.Delta) and name in v.fresh
    ]
    assert len(delta_terms) == 1
    return _get_support_value(delta_terms[0], name, **kwargs)


@_get_support_value.register(funsor.delta.Delta)
def _get_support_value_delta(funsor_dist, name, **kwargs):
    assert name in funsor_dist.fresh
    return OrderedDict(funsor_dist.terms)[name][0]


@_get_support_value.register(funsor.Tensor)
def _get_support_value_tensor(funsor_dist, name, **kwargs):
    assert name in funsor_dist.inputs
    return funsor.Tensor(
        funsor.ops.new_arange(funsor_dist.data, funsor_dist.inputs[name].size),
        OrderedDict([(name, funsor_dist.inputs[name])]),
        funsor_dist.inputs[name].size,
    )


@_get_support_value.register(funsor.distribution.Distribution)
def _get_support_value_distribution(funsor_dist, name, expand=False):
    assert name == funsor_dist.value.name
    return funsor_dist.enumerate_support(expand=expand)


def _enum_strategy_default(dist, msg):
    sample_inputs = OrderedDict(
        (f.name, funsor.Bint[f.size])
        for f in msg["cond_indep_stack"]
        if f.vectorized and f.name not in dist.inputs
    )
    sampled_dist = dist.sample(msg["name"], sample_inputs)
    sampled_dist -= sum([math.log(v.size) for v in sample_inputs.values()], 0)
    return sampled_dist


def _enum_strategy_diagonal(dist, msg):
    sample_dim_name = "{}__PARTICLES".format(msg["name"])
    sample_inputs = OrderedDict(
        {sample_dim_name: funsor.Bint[msg["infer"]["num_samples"]]}
    )
    plate_names = frozenset(f.name for f in msg["cond_indep_stack"] if f.vectorized)
    ancestor_names = frozenset(
        k
        for k, v in dist.inputs.items()
        if v.dtype != "real" and k != msg["name"] and k not in plate_names
    )
    # TODO should the ancestor_indices be pyro.observed?
    ancestor_indices = {name: sample_dim_name for name in ancestor_names}
    denom = (
        sum([math.log(v.size) for v in sample_inputs.values()], 0)
        if not ancestor_indices
        else math.log(msg["infer"]["num_samples"])
    )
    sampled_dist = dist(**ancestor_indices).sample(
        msg["name"], sample_inputs if not ancestor_indices else None
    )
    sampled_dist -= denom
    return sampled_dist


def _enum_strategy_mixture(dist, msg):
    sample_dim_name = "{}__PARTICLES".format(msg["name"])
    sample_inputs = OrderedDict(
        {sample_dim_name: funsor.Bint[msg["infer"]["num_samples"]]}
    )
    plate_names = frozenset(f.name for f in msg["cond_indep_stack"] if f.vectorized)
    ancestor_names = frozenset(
        k
        for k, v in dist.inputs.items()
        if v.dtype != "real" and k != msg["name"] and k not in plate_names
    )
    plate_inputs = OrderedDict((k, dist.inputs[k]) for k in plate_names)
    # TODO should the ancestor_indices be pyro.sampled?
    ancestor_indices = {
        # TODO make this comprehension less gross
        name: _get_support_value(
            funsor.torch.distributions.CategoricalLogits(
                # sample different ancestors for each plate slice
                logits=funsor.Tensor(
                    # TODO avoid use of torch.zeros here in favor of funsor.ops.new_zeros
                    torch.zeros((1,)).expand(
                        tuple(v.dtype for v in plate_inputs.values())
                        + (dist.inputs[name].dtype,)
                    ),
                    plate_inputs,
                ),
            )(value=name).sample(name, sample_inputs),
            name,
        )
        for name in ancestor_names
    }
    denom = (
        sum([math.log(v.size) for v in sample_inputs.values()], 0)
        if not ancestor_indices
        else math.log(msg["infer"]["num_samples"])
    )
    sampled_dist = dist(**ancestor_indices).sample(
        msg["name"], sample_inputs if not ancestor_indices else None
    )
    sampled_dist -= denom
    return sampled_dist


def _enum_strategy_full(dist, msg):
    sample_dim_name = "{}__PARTICLES".format(msg["name"])
    sample_inputs = OrderedDict(
        {sample_dim_name: funsor.Bint[msg["infer"]["num_samples"]]}
    )
    sampled_dist = dist.sample(msg["name"], sample_inputs)
    sampled_dist -= math.log(msg["infer"]["num_samples"])
    return sampled_dist


def _enum_strategy_exact(dist, msg):
    if isinstance(dist, funsor.Tensor):
        dist = dist - dist.reduce(funsor.ops.logaddexp, msg["name"])
    return dist


def enumerate_site(dist, msg):
    # TODO come up with a better dispatch system for enumeration strategies
    if msg["infer"]["enumerate"] == "flat":
        return _enum_strategy_default(dist, msg)
    elif msg["infer"].get("num_samples", None) is None:
        return _enum_strategy_exact(dist, msg)
    elif msg["infer"]["num_samples"] > 1 and (
        msg["infer"].get("expand", False) or msg["infer"].get("tmc") == "full"
    ):
        return _enum_strategy_full(dist, msg)
    elif (
        msg["infer"]["num_samples"] > 1
        and msg["infer"].get("tmc", "diagonal") == "diagonal"
    ):
        return _enum_strategy_diagonal(dist, msg)
    elif msg["infer"]["num_samples"] > 1 and msg["infer"]["tmc"] == "mixture":
        return _enum_strategy_mixture(dist, msg)
    raise ValueError("{} not valid enum strategy".format(msg))


class EnumMessenger(NamedMessenger):
    """
    This version of :class:`~EnumMessenger` uses :func:`~pyro.contrib.funsor.to_data`
    to allocate a fresh enumeration dim for each discrete sample site.
    """

    def _pyro_sample(self, msg):
        if (
            msg["done"]
            or msg["is_observed"]
            or msg["infer"].get("enumerate") not in {"flat", "parallel"}
            or isinstance(msg["fn"], _Subsample)
        ):
            return

        if "funsor" not in msg:
            msg["funsor"] = {}

        unsampled_log_measure = to_funsor(msg["fn"], output=funsor.Real)(
            value=msg["name"]
        )
        msg["funsor"]["log_measure"] = enumerate_site(unsampled_log_measure, msg)
        msg["funsor"]["value"] = _get_support_value(
            msg["funsor"]["log_measure"],
            msg["name"],
            expand=msg["infer"].get("expand", False),
        )
        msg["value"] = to_data(msg["funsor"]["value"])
        msg["done"] = True


def queue(
    fn=None,
    queue=None,
    max_tries=int(1e6),
    num_samples=-1,
    extend_fn=pyro.poutine.util.enum_extend,
    escape_fn=pyro.poutine.util.discrete_escape,
):
    """
    Used in sequential enumeration over discrete variables (copied from poutine.queue).

    Given a stochastic function and a queue,
    return a return value from a complete trace in the queue.

    :param fn: a stochastic function (callable containing Pyro primitive calls)
    :param q: a queue data structure like multiprocessing.Queue to hold partial traces
    :param max_tries: maximum number of attempts to compute a single complete trace
    :param extend_fn: function (possibly stochastic) that takes a partial trace and a site,
        and returns a list of extended traces
    :param escape_fn: function (possibly stochastic) that takes a partial trace and a site,
        and returns a boolean value to decide whether to exit
    :param num_samples: optional number of extended traces for extend_fn to return
    :returns: stochastic function decorated with poutine logic
    """

    # TODO rewrite this to use purpose-built trace/replay handlers
    def wrapper(wrapped):
        def _fn(*args, **kwargs):
            for i in range(max_tries):
                assert (
                    not queue.empty()
                ), "trying to get() from an empty queue will deadlock"

                next_trace = queue.get()
                try:
                    ftr = TraceMessenger()(
                        EscapeMessenger(
                            escape_fn=functools.partial(escape_fn, next_trace)
                        )(ReplayMessenger(trace=next_trace)(wrapped))
                    )
                    return ftr(*args, **kwargs)
                except pyro.poutine.runtime.NonlocalExit as site_container:
                    site_container.reset_stack()  # TODO implement missing ._reset()s
                    for tr in extend_fn(
                        ftr.trace.copy(), site_container.site, num_samples=num_samples
                    ):
                        queue.put(tr)

            raise ValueError("max tries ({}) exceeded".format(str(max_tries)))

        return _fn

    return wrapper(fn) if fn is not None else wrapper



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import torch

import pyro
import pyro.distributions as dist
import pyro.poutine as poutine

from .reparam import Reparam


class SplitReparam(Reparam):
    """
    Reparameterizer to split a random variable along a dimension, similar to
    :func:`torch.split`.

    This is useful for treating different parts of a tensor with different
    reparameterizers or inference methods. For example when performing HMC
    inference on a time series, you can first apply
    :class:`~pyro.infer.reparam.discrete_cosine.DiscreteCosineReparam` or
    :class:`~pyro.infer.reparam.haar.HaarReparam`, then apply
    :class:`SplitReparam` to split into low-frequency and high-frequency
    components, and finally add the low-frequency components to the
    ``full_mass`` matrix together with globals.

    :param sections: Size of a single chunk or list of sizes for
        each chunk.
    :type: list(int)
    :param int dim: Dimension along which to split. Defaults to -1.
    """

    def __init__(self, sections, dim):
        assert isinstance(dim, int) and dim < 0
        assert isinstance(sections, list)
        assert all(isinstance(size, int) for size in sections)
        self.event_dim = -dim
        self.sections = sections

    def apply(self, msg):
        name = msg["name"]
        fn = msg["fn"]
        value = msg["value"]
        is_observed = msg["is_observed"]
        assert fn.event_dim >= self.event_dim

        # Split value into parts.
        value_split = [None] * len(self.sections)
        if value is not None:
            value_split[:] = value.split(self.sections, -self.event_dim)

        # Draw independent parts.
        dim = fn.event_dim - self.event_dim
        left_shape = fn.event_shape[:dim]
        right_shape = fn.event_shape[1 + dim :]
        for i, size in enumerate(self.sections):
            event_shape = left_shape + (size,) + right_shape
            value_split[i] = pyro.sample(
                f"{name}_split_{i}",
                dist.ImproperUniform(fn.support, fn.batch_shape, event_shape),
                obs=value_split[i],
                infer={"is_observed": is_observed},
            )

        # Combine parts into value.
        if value is None:
            value = torch.cat(value_split, dim=-self.event_dim)

        if poutine.get_mask() is False:
            log_density = 0.0
        else:
            log_density = fn.log_prob(value)
        new_fn = dist.Delta(value, event_dim=fn.event_dim, log_density=log_density)
        return {"fn": new_fn, "value": value, "is_observed": True}



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import inspect
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Type,
    Union,
    ValuesView,
)

import torch
from torch import Tensor
from torch.nn.utils import clip_grad_norm_, clip_grad_value_
from torch.optim import Optimizer

import pyro
from pyro.optim.adagrad_rmsprop import AdagradRMSProp as pt_AdagradRMSProp
from pyro.optim.clipped_adam import ClippedAdam as pt_ClippedAdam
from pyro.optim.dct_adam import DCTAdam as pt_DCTAdam
from pyro.params.param_store import (
    module_from_param_with_module_name,
    normalize_param_name,
    user_param_name,
)


def is_scheduler(optimizer) -> bool:
    """
    Helper method to determine whether a PyTorch object is either a PyTorch
    optimizer (return false) or a optimizer wrapped in an LRScheduler e.g. a
    ``ReduceLROnPlateau`` or subclasses of ``_LRScheduler`` (return true).
    """
    # This uses duck typing rather than isinstance() because (1) PyTorch
    # provides no comprehensive class hierarchy, and (2) the base class
    # _LRScheduler of the majority of schedulers is private.
    return hasattr(optimizer, "optimizer")


def _get_state_dict(optimizer) -> dict:
    """
    Helper to get the state dict for either a raw optimizer or an optimizer
    wrapped in an LRScheduler.
    """
    if is_scheduler(optimizer):
        state = {
            "scheduler": optimizer.state_dict(),
            "optimizer": optimizer.optimizer.state_dict(),
        }
    else:
        state = optimizer.state_dict()
    return state


def _load_state_dict(optimizer, state: dict) -> None:
    """
    Helper to load the state dict into either a raw optimizer or an optimizer
    wrapped in an LRScheduler.
    """
    if is_scheduler(optimizer):
        optimizer.load_state_dict(state["scheduler"])
        optimizer.optimizer.load_state_dict(state["optimizer"])
    else:
        optimizer.load_state_dict(state)


class PyroOptim:
    """
    A wrapper for torch.optim.Optimizer objects that helps with managing dynamically generated parameters.

    :param optim_constructor: a torch.optim.Optimizer
    :param optim_args: a dictionary of learning arguments for the optimizer or a callable that returns
        such dictionaries
    :param clip_args: a dictionary of clip_norm and/or clip_value args or a callable that returns
        such dictionaries
    """

    def __init__(
        self,
        optim_constructor: Union[Callable, Optimizer, Type[Optimizer]],
        optim_args: Union[Dict, Callable[..., Dict]],
        clip_args: Optional[Union[Dict, Callable[..., Dict]]] = None,
    ):
        self.pt_optim_constructor = optim_constructor

        # must be callable or dict
        assert callable(optim_args) or isinstance(
            optim_args, dict
        ), "optim_args must be function that returns defaults or a defaults dictionary"

        if clip_args is None:
            clip_args = {}

        # must be callable or dict
        assert callable(clip_args) or isinstance(
            clip_args, dict
        ), "clip_args must be function that returns defaults or a defaults dictionary"

        # hold our args to be called/used
        self.pt_optim_args = optim_args
        if callable(optim_args):
            self.pt_optim_args_argc = len(inspect.signature(optim_args).parameters)
        self.pt_clip_args = clip_args

        # holds the torch optimizer objects
        self.optim_objs: Dict = {}
        self.grad_clip: Dict = {}

        # any optimizer state that's waiting to be consumed (because that parameter hasn't been seen before)
        self._state_waiting_to_be_consumed: Dict = {}

    def __call__(self, params: Union[List, ValuesView], *args, **kwargs) -> None:
        """
        :param params: a list of parameters
        :type params: an iterable of strings

        Do an optimization step for each param in params. If a given param has never been seen before,
        initialize an optimizer for it.
        """
        for p in params:
            # if we have not seen this param before, we instantiate an optim object to deal with it
            if p not in self.optim_objs:
                # create a single optim object for that param
                optimizer = self.optim_objs[p] = self._get_optim(p)
                # create a gradient clipping function if specified
                self.grad_clip[p] = self._get_grad_clip(p)
                # set state from _state_waiting_to_be_consumed if present
                param_name = pyro.get_param_store().param_name(p)
                state = self._state_waiting_to_be_consumed.pop(param_name, None)
                if state is not None:
                    _load_state_dict(optimizer, state)

            if self.grad_clip[p] is not None:
                self.grad_clip[p](p)

            if (
                hasattr(torch.optim.lr_scheduler, "_LRScheduler")
                and isinstance(
                    self.optim_objs[p], torch.optim.lr_scheduler._LRScheduler
                )
                or hasattr(torch.optim.lr_scheduler, "LRScheduler")
                and isinstance(self.optim_objs[p], torch.optim.lr_scheduler.LRScheduler)
                or isinstance(
                    self.optim_objs[p], torch.optim.lr_scheduler.ReduceLROnPlateau
                )
            ):
                # if optim object was a scheduler, perform an optimizer step
                self.optim_objs[p].optimizer.step(*args, **kwargs)
            else:
                self.optim_objs[p].step(*args, **kwargs)

    def get_state(self) -> Dict:
        """
        Get state associated with all the optimizers in the form of a dictionary with
        key-value pairs (parameter name, optim state dicts)
        """
        state_dict = {}
        for param in self.optim_objs:
            param_name = pyro.get_param_store().param_name(param)
            state_dict[param_name] = _get_state_dict(self.optim_objs[param])
        return state_dict

    def set_state(self, state_dict: Dict) -> None:
        """
        Set the state associated with all the optimizers using the state obtained
        from a previous call to get_state()
        """
        self._state_waiting_to_be_consumed.update(state_dict)

    def save(self, filename: str) -> None:
        """
        :param filename: file name to save to
        :type filename: str

        Save optimizer state to disk
        """
        with open(filename, "wb") as output_file:
            torch.save(self.get_state(), output_file)

    def load(self, filename: str, map_location=None) -> None:
        """
        :param filename: file name to load from
        :type filename: str
        :param map_location: torch.load() map_location parameter
        :type map_location: function, torch.device, string or a dict

        Load optimizer state from disk
        """
        with open(filename, "rb") as input_file:
            state = torch.load(input_file, map_location=map_location)
        self.set_state(state)

    def _get_optim(self, param: Union[Iterable[Tensor], Iterable[Dict[Any, Any]]]):
        return self.pt_optim_constructor([param], **self._get_optim_args(param))  # type: ignore

    # helper to fetch the optim args if callable (only used internally)
    def _get_optim_args(self, param: Union[Iterable[Tensor], Iterable[Dict]]):
        # If we were passed a function, we call function with a
        # fully qualified name e.g. 'mymodule.mysubmodule.bias'.
        if callable(self.pt_optim_args):
            param_name = pyro.get_param_store().param_name(param)
            if self.pt_optim_args_argc == 1:
                # Normalize to the format of nn.Module.named_parameters().
                normal_name = normalize_param_name(param_name)
                opt_dict = self.pt_optim_args(normal_name)
            else:
                # DEPRECATED Split param name in to pieces.
                module_name = module_from_param_with_module_name(param_name)
                stripped_param_name = user_param_name(param_name)
                opt_dict = self.pt_optim_args(module_name, stripped_param_name)

            # must be dictionary
            assert isinstance(
                opt_dict, dict
            ), "per-param optim arg must return defaults dictionary"
            return opt_dict
        else:
            return self.pt_optim_args

    def _get_grad_clip(self, param: str):
        grad_clip_args = self._get_grad_clip_args(param)

        if not grad_clip_args:
            return None

        def _clip_grad(params: Union[Tensor, Iterable[Tensor]]):
            self._clip_grad(params, **grad_clip_args)

        return _clip_grad

    def _get_grad_clip_args(self, param: str) -> Dict:
        # if we were passed a fct, we call fct with param info
        # arguments are (module name, param name) e.g. ('mymodule', 'bias')
        if callable(self.pt_clip_args):
            # get param name
            param_name = pyro.get_param_store().param_name(param)
            module_name = module_from_param_with_module_name(param_name)
            stripped_param_name = user_param_name(param_name)

            # invoke the user-provided callable
            clip_dict = self.pt_clip_args(module_name, stripped_param_name)

            # must be dictionary
            assert isinstance(
                clip_dict, dict
            ), "per-param clip arg must return defaults dictionary"
            return clip_dict
        else:
            return self.pt_clip_args

    @staticmethod
    def _clip_grad(
        params: Union[Tensor, Iterable[Tensor]],
        clip_norm: Optional[Union[int, float]] = None,
        clip_value: Optional[Union[int, float]] = None,
    ) -> None:
        if clip_norm is not None:
            clip_grad_norm_(params, clip_norm)
        if clip_value is not None:
            clip_grad_value_(params, clip_value)


def AdagradRMSProp(optim_args: Dict) -> PyroOptim:
    """
    Wraps :class:`pyro.optim.adagrad_rmsprop.AdagradRMSProp` with :class:`~pyro.optim.optim.PyroOptim`.
    """
    return PyroOptim(pt_AdagradRMSProp, optim_args)


def ClippedAdam(optim_args: Dict) -> PyroOptim:
    """
    Wraps :class:`pyro.optim.clipped_adam.ClippedAdam` with :class:`~pyro.optim.optim.PyroOptim`.
    """
    return PyroOptim(pt_ClippedAdam, optim_args)


def DCTAdam(optim_args: Dict) -> PyroOptim:
    """
    Wraps :class:`pyro.optim.dct_adam.DCTAdam` with :class:`~pyro.optim.optim.PyroOptim`.
    """
    return PyroOptim(pt_DCTAdam, optim_args)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import torch
import torch.nn as nn


class Profile(nn.Module):
    """
    Profile HMM state arrangement. Parameterizes an HMM according to
    Equation S40 in [1] (with r_{M+1,j} = 1 and u_{M+1,j} = 0
    for j in {0, 1, 2}). For further background on profile HMMs see [2].

    **References**

    [1] E. N. Weinstein, D. S. Marks (2021)
    "Generative probabilistic biological sequence models that account for
    mutational variability"
    https://www.biorxiv.org/content/10.1101/2020.07.31.231381v2.full.pdf

    [2] R. Durbin, S. R. Eddy, A. Krogh, and G. Mitchison (1998)
    "Biological sequence analysis: probabilistic models of proteins and nucleic
    acids"
    Cambridge university press

    :param M: Length of regressor sequence.
    :type M: int
    :param epsilon: A small value for numerical stability.
    :type epsilon: float
    """

    def __init__(self, M, epsilon=1e-32):
        super().__init__()
        self.M = M
        self.K = 2 * M + 1
        self.epsilon = epsilon

        self._make_transfer()

    def _make_transfer(self):
        """Set up linear transformations (transfer matrices) for converting
        from profile HMM parameters to standard HMM parameters."""
        M, K = self.M, self.K

        # Overview:
        # r -> insertion parameters
        # u -> deletion parameters
        # indices: m in {0, ..., M} and j in {0, 1, 2}; final index corresponds
        # to simplex dimensions, i.e. 1 - r and r (in that order)
        # null -> locations in the transition matrix equal to 0
        # ...transf_0 -> initial transition vector
        # ...transf -> transition matrix
        # We fix r_{M+1,j} = 1 for j in {0, 1, 2}
        self.register_buffer("r_transf_0", torch.zeros((M, 3, 2, K)))
        self.register_buffer("u_transf_0", torch.zeros((M, 3, 2, K)))
        self.register_buffer("null_transf_0", torch.zeros((K,)))
        m, g = -1, 0
        for gp in range(2):
            for mp in range(M + gp):
                kp = mg2k(mp, gp, M)
                if m + 1 - g == mp and gp == 0:
                    self.r_transf_0[m + 1 - g, g, 0, kp] = 1
                    self.u_transf_0[m + 1 - g, g, 0, kp] = 1

                elif m + 1 - g < mp and gp == 0:
                    self.r_transf_0[m + 1 - g, g, 0, kp] = 1
                    self.u_transf_0[m + 1 - g, g, 1, kp] = 1
                    for mpp in range(m + 2 - g, mp):
                        self.r_transf_0[mpp, 2, 0, kp] = 1
                        self.u_transf_0[mpp, 2, 1, kp] = 1
                    self.r_transf_0[mp, 2, 0, kp] = 1
                    self.u_transf_0[mp, 2, 0, kp] = 1

                elif m + 1 - g == mp and gp == 1:
                    if mp < M:
                        self.r_transf_0[m + 1 - g, g, 1, kp] = 1

                elif m + 1 - g < mp and gp == 1:
                    self.r_transf_0[m + 1 - g, g, 0, kp] = 1
                    self.u_transf_0[m + 1 - g, g, 1, kp] = 1
                    for mpp in range(m + 2 - g, mp):
                        self.r_transf_0[mpp, 2, 0, kp] = 1
                        self.u_transf_0[mpp, 2, 1, kp] = 1
                    if mp < M:
                        self.r_transf_0[mp, 2, 1, kp] = 1

                else:
                    self.null_transf_0[kp] = 1

        self.register_buffer("r_transf", torch.zeros((M, 3, 2, K, K)))
        self.register_buffer("u_transf", torch.zeros((M, 3, 2, K, K)))
        self.register_buffer("null_transf", torch.zeros((K, K)))
        for g in range(2):
            for m in range(M + g):
                for gp in range(2):
                    for mp in range(M + gp):
                        k, kp = mg2k(m, g, M), mg2k(mp, gp, M)
                        if m + 1 - g == mp and gp == 0:
                            self.r_transf[m + 1 - g, g, 0, k, kp] = 1
                            self.u_transf[m + 1 - g, g, 0, k, kp] = 1

                        elif m + 1 - g < mp and gp == 0:
                            self.r_transf[m + 1 - g, g, 0, k, kp] = 1
                            self.u_transf[m + 1 - g, g, 1, k, kp] = 1
                            self.r_transf[(m + 2 - g) : mp, 2, 0, k, kp] = 1
                            self.u_transf[(m + 2 - g) : mp, 2, 1, k, kp] = 1
                            self.r_transf[mp, 2, 0, k, kp] = 1
                            self.u_transf[mp, 2, 0, k, kp] = 1

                        elif m + 1 - g == mp and gp == 1:
                            if mp < M:
                                self.r_transf[m + 1 - g, g, 1, k, kp] = 1

                        elif m + 1 - g < mp and gp == 1:
                            self.r_transf[m + 1 - g, g, 0, k, kp] = 1
                            self.u_transf[m + 1 - g, g, 1, k, kp] = 1
                            self.r_transf[(m + 2 - g) : mp, 2, 0, k, kp] = 1
                            self.u_transf[(m + 2 - g) : mp, 2, 1, k, kp] = 1
                            if mp < M:
                                self.r_transf[mp, 2, 1, k, kp] = 1

                        else:
                            self.null_transf[k, kp] = 1

        self.register_buffer("vx_transf", torch.zeros((M, K)))
        self.register_buffer("vc_transf", torch.zeros((M + 1, K)))
        for g in range(2):
            for m in range(M + g):
                k = mg2k(m, g, M)
                if g == 0:
                    self.vx_transf[m, k] = 1
                elif g == 1:
                    self.vc_transf[m, k] = 1

    def forward(
        self,
        precursor_seq_logits,
        insert_seq_logits,
        insert_logits,
        delete_logits,
        substitute_logits=None,
    ):
        """
        Assemble HMM parameters given profile parameters.

        :param ~torch.Tensor precursor_seq_logits: Regressor sequence
            *log(x)*. Should have rightmost dimension ``(M, D)`` and be
            broadcastable to ``(batch_size, M, D)``, where
            D is the latent alphabet size. Should be normalized to one along the
            final axis, i.e. ``precursor_seq_logits.logsumexp(-1) = zeros``.
        :param ~torch.Tensor insert_seq_logits: Insertion sequence *log(c)*.
            Should have rightmost dimension ``(M+1, D)`` and be broadcastable
            to ``(batch_size, M+1, D)``. Should be normalized
            along the final axis.
        :param ~torch.Tensor insert_logits: Insertion probabilities *log(r)*.
            Should have rightmost dimension ``(M, 3, 2)`` and be broadcastable
            to ``(batch_size, M, 3, 2)``. Should be normalized along the
            final axis.
        :param ~torch.Tensor delete_logits: Deletion probabilities *log(u)*.
            Should have rightmost dimension ``(M, 3, 2)`` and be broadcastable
            to ``(batch_size, M, 3, 2)``. Should be normalized along the
            final axis.
        :param ~torch.Tensor substitute_logits: Substitution probabilities
            *log(l)*. Should have rightmost dimension ``(D, B)``, where
            B is the alphabet size of the data, and broadcastable to
            ``(batch_size, D, B)``. Must be normalized along the
            final axis.
        :return: *initial_logits*, *transition_logits*, and
            *observation_logits*. These parameters can be used to directly
            initialize the MissingDataDiscreteHMM distribution.
        :rtype: ~torch.Tensor, ~torch.Tensor, ~torch.Tensor
        """
        initial_logits = (
            torch.einsum("...ijk,ijkl->...l", delete_logits, self.u_transf_0)
            + torch.einsum("...ijk,ijkl->...l", insert_logits, self.r_transf_0)
            + (-1 / self.epsilon) * self.null_transf_0
        )
        transition_logits = (
            torch.einsum("...ijk,ijklf->...lf", delete_logits, self.u_transf)
            + torch.einsum("...ijk,ijklf->...lf", insert_logits, self.r_transf)
            + (-1 / self.epsilon) * self.null_transf
        )
        # Broadcasting for concatenation.
        if len(precursor_seq_logits.size()) > len(insert_seq_logits.size()):
            insert_seq_logits = insert_seq_logits.unsqueeze(0).expand(
                [precursor_seq_logits.size()[0], -1, -1]
            )
        elif len(insert_seq_logits.size()) > len(precursor_seq_logits.size()):
            precursor_seq_logits = precursor_seq_logits.unsqueeze(0).expand(
                [insert_seq_logits.size()[0], -1, -1]
            )
        seq_logits = torch.cat([precursor_seq_logits, insert_seq_logits], dim=-2)

        # Option to include the substitution matrix.
        if substitute_logits is not None:
            observation_logits = torch.logsumexp(
                seq_logits.unsqueeze(-1) + substitute_logits.unsqueeze(-3), dim=-2
            )
        else:
            observation_logits = seq_logits

        return initial_logits, transition_logits, observation_logits


def mg2k(m, g, M):
    """Convert from (m, g) indexing to k indexing."""
    return m + M * g



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import torch
from torch.distributions.transforms import Transform

from .. import constraints
from ..util import copy_docs_from


@copy_docs_from(Transform)
class LowerCholeskyAffine(Transform):
    """
    A bijection of the form,

        :math:`\\mathbf{y} = \\mathbf{L} \\mathbf{x} + \\mathbf{r}`

    where `\\mathbf{L}` is a lower triangular matrix and `\\mathbf{r}` is a vector.

    :param loc: the fixed D-dimensional vector to shift the input by.
    :type loc: torch.tensor
    :param scale_tril: the D x D lower triangular matrix used in the transformation.
    :type scale_tril: torch.tensor

    """

    domain = constraints.real_vector
    codomain = constraints.real_vector
    bijective = True
    volume_preserving = False

    def __init__(self, loc, scale_tril, cache_size=0):
        super().__init__(cache_size=cache_size)
        self.loc = loc
        self.scale_tril = scale_tril
        assert (
            loc.size(-1) == scale_tril.size(-1) == scale_tril.size(-2)
        ), "loc and scale_tril must be of size D and D x D, respectively (instead: {}, {})".format(
            loc.shape, scale_tril.shape
        )

    def _call(self, x):
        """
        :param x: the input into the bijection
        :type x: torch.Tensor

        Invokes the bijection x=>y; in the prototypical context of a
        :class:`~pyro.distributions.TransformedDistribution` `x` is a sample from
        the base distribution (or the output of a previous transform)
        """
        return torch.matmul(self.scale_tril, x.unsqueeze(-1)).squeeze(-1) + self.loc

    def _inverse(self, y):
        """
        :param y: the output of the bijection
        :type y: torch.Tensor

        Inverts y => x.
        """
        return torch.linalg.solve_triangular(
            self.scale_tril, (y - self.loc).unsqueeze(-1), upper=False
        ).squeeze(-1)

    def log_abs_det_jacobian(self, x, y):
        """
        Calculates the elementwise determinant of the log Jacobian, i.e.
        log(abs(dy/dx)).
        """
        return (
            torch.ones(x.size()[:-1], dtype=x.dtype, layout=x.layout, device=x.device)
            * self.scale_tril.diag().log().sum()
        )

    def with_cache(self, cache_size=1):
        if self._cache_size == cache_size:
            return self
        return LowerCholeskyAffine(self.loc, self.scale_tril, cache_size=cache_size)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import pyro.distributions as dist

from .reparam import Reparam


class LinearHMMReparam(Reparam):
    """
    Auxiliary variable reparameterizer for
    :class:`~pyro.distributions.LinearHMM` random variables.

    This defers to component reparameterizers to create auxiliary random
    variables conditioned on which the process becomes a
    :class:`~pyro.distributions.GaussianHMM` . If the ``observation_dist`` is a
    :class:`~pyro.distributions.TransformedDistribution` this reorders those
    transforms so that the result is a
    :class:`~pyro.distributions.TransformedDistribution` of
    :class:`~pyro.distributions.GaussianHMM` .

    This is useful for training the parameters of a
    :class:`~pyro.distributions.LinearHMM` distribution, whose
    :meth:`~pyro.distributions.LinearHMM.log_prob` method is undefined.  To
    perform inference in the presence of non-Gaussian factors such as
    :meth:`~pyro.distributions.Stable`, :meth:`~pyro.distributions.StudentT` or
    :meth:`~pyro.distributions.LogNormal` , configure with
    :class:`~pyro.infer.reparam.studentt.StudentTReparam` ,
    :class:`~pyro.infer.reparam.stable.StableReparam` ,
    :class:`~pyro.infer.reparam.stable.SymmetricStableReparam` , etc.  component
    reparameterizers for ``init``, ``trans``, and ``scale``. For example::

        hmm = LinearHMM(
            init_dist=Stable(1,0,1,0).expand([2]).to_event(1),
            trans_matrix=torch.eye(2),
            trans_dist=MultivariateNormal(torch.zeros(2), torch.eye(2)),
            obs_matrix=torch.eye(2),
            obs_dist=TransformedDistribution(
                Stable(1.5,-0.5,1.0).expand([2]).to_event(1),
                ExpTransform()))

        rep = LinearHMMReparam(init=SymmetricStableReparam(),
                               obs=StableReparam())

        with poutine.reparam(config={"hmm": rep}):
            pyro.sample("hmm", hmm, obs=data)

    :param init: Optional reparameterizer for the initial distribution.
    :type init: ~pyro.infer.reparam.reparam.Reparam
    :param trans: Optional reparameterizer for the transition distribution.
    :type trans: ~pyro.infer.reparam.reparam.Reparam
    :param obs: Optional reparameterizer for the observation distribution.
    :type obs: ~pyro.infer.reparam.reparam.Reparam
    """

    def __init__(self, init=None, trans=None, obs=None):
        assert init is None or isinstance(init, Reparam)
        assert trans is None or isinstance(trans, Reparam)
        assert obs is None or isinstance(obs, Reparam)
        self.init = init
        self.trans = trans
        self.obs = obs

    def apply(self, msg):
        name = msg["name"]
        fn = msg["fn"]
        value = msg["value"]
        is_observed = msg["is_observed"]

        fn, event_dim = self._unwrap(fn)
        assert isinstance(fn, (dist.LinearHMM, dist.IndependentHMM))
        if fn.duration is None:
            raise ValueError(
                "LinearHMMReparam requires duration to be specified "
                "on targeted LinearHMM distributions"
            )

        # Unwrap IndependentHMM.
        if isinstance(fn, dist.IndependentHMM):
            indep_value = None
            if value is not None:
                indep_value = value.transpose(-1, -2).unsqueeze(-1)
            msg = self.apply(
                {
                    "name": name,
                    "fn": fn.base_dist.to_event(1),
                    "value": indep_value,
                    "is_observed": is_observed,
                }
            )
            hmm = msg["fn"]
            hmm = dist.IndependentHMM(hmm.to_event(-1))
            if msg["value"] is not indep_value:
                value = msg["value"].squeeze(-1).transpose(-1, -2)
            return {"fn": hmm, "value": value, "is_observed": is_observed}

        # Reparameterize the initial distribution as conditionally Gaussian.
        init_dist = fn.initial_dist
        if self.init is not None:
            msg = self.init.apply(
                {
                    "name": f"{name}_init",
                    "fn": self._wrap(init_dist, event_dim - 1),
                    "value": None,
                    "is_observed": False,
                }
            )
            init_dist = msg["fn"]
            init_dist = init_dist.to_event(1 - init_dist.event_dim)

        # Reparameterize the transition distribution as conditionally Gaussian.
        trans_dist = fn.transition_dist
        if self.trans is not None:
            if trans_dist.batch_shape[-1] != fn.duration:
                trans_dist = trans_dist.expand(
                    trans_dist.batch_shape[:-1] + (fn.duration,)
                )
            msg = self.trans.apply(
                {
                    "name": f"{name}_trans",
                    "fn": self._wrap(trans_dist, event_dim),
                    "value": None,
                    "is_observed": False,
                }
            )
            trans_dist = msg["fn"]
            trans_dist = trans_dist.to_event(1 - trans_dist.event_dim)

        # Reparameterize the observation distribution as conditionally Gaussian.
        obs_dist = fn.observation_dist
        if self.obs is not None:
            if obs_dist.batch_shape[-1] != fn.duration:
                obs_dist = obs_dist.expand(obs_dist.batch_shape[:-1] + (fn.duration,))
            msg = self.obs.apply(
                {
                    "name": f"{name}_obs",
                    "fn": self._wrap(obs_dist, event_dim),
                    "value": value,
                    "is_observed": is_observed,
                }
            )
            obs_dist = msg["fn"]
            obs_dist = obs_dist.to_event(1 - obs_dist.event_dim)
            value = msg["value"]
            is_observed = msg["is_observed"]

        # Reparameterize the entire HMM as conditionally Gaussian.
        hmm = dist.GaussianHMM(
            init_dist,
            fn.transition_matrix,
            trans_dist,
            fn.observation_matrix,
            obs_dist,
            duration=fn.duration,
        )
        hmm = self._wrap(hmm, event_dim)

        # Apply any observation transforms.
        if fn.transforms:
            hmm = dist.TransformedDistribution(hmm, fn.transforms)

        return {"fn": hmm, "value": value, "is_observed": is_observed}



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import queue
import warnings

import torch

import pyro.poutine as poutine
from pyro.distributions.util import is_identically_zero
from pyro.infer.elbo import ELBO
from pyro.infer.enum import (
    get_importance_trace,
    iter_discrete_escape,
    iter_discrete_extend,
)
from pyro.infer.util import compute_site_dice_factor, is_validation_enabled, torch_item
from pyro.ops import packed
from pyro.ops.contract import einsum
from pyro.poutine.enum_messenger import EnumMessenger
from pyro.util import check_traceenum_requirements, warn_if_nan


def _compute_dice_factors(model_trace, guide_trace):
    """
    compute per-site DiCE log-factors for non-reparameterized proposal sites
    this logic is adapted from pyro.infer.util.Dice.__init__
    """
    log_probs = []
    for role, trace in zip(("model", "guide"), (model_trace, guide_trace)):
        for name, site in trace.nodes.items():
            if site["type"] != "sample" or site["is_observed"]:
                continue
            if role == "model" and name in guide_trace:
                continue

            log_prob, log_denom = compute_site_dice_factor(site)
            if not is_identically_zero(log_denom):
                dims = log_prob._pyro_dims
                log_prob = log_prob - log_denom
                log_prob._pyro_dims = dims
            if not is_identically_zero(log_prob):
                log_probs.append(log_prob)

    return log_probs


def _compute_tmc_factors(model_trace, guide_trace):
    """
    compute per-site log-factors for all observed and unobserved variables
    log-factors are log(p / q) for unobserved sites and log(p) for observed sites
    """
    log_factors = []
    for name, site in guide_trace.nodes.items():
        if site["type"] != "sample" or site["is_observed"]:
            continue
        log_proposal = site["packed"]["log_prob"]
        log_factors.append(packed.neg(log_proposal))
    for name, site in model_trace.nodes.items():
        if site["type"] != "sample":
            continue
        if (
            site["name"] not in guide_trace
            and not site["is_observed"]
            and site["infer"].get("enumerate", None) == "parallel"
            and site["infer"].get("num_samples", -1) > 0
        ):
            # site was sampled from the prior
            log_proposal = packed.neg(site["packed"]["log_prob"])
            log_factors.append(log_proposal)
        log_factors.append(site["packed"]["log_prob"])
    return log_factors


def _compute_tmc_estimate(model_trace, guide_trace):
    """
    Use :func:`~pyro.ops.contract.einsum` to compute the Tensor Monte Carlo
    estimate of the marginal likelihood given parallel-sampled traces.
    """
    # factors
    log_factors = _compute_tmc_factors(model_trace, guide_trace)
    log_factors += _compute_dice_factors(model_trace, guide_trace)

    if not log_factors:
        return 0.0

    # loss
    eqn = ",".join([f._pyro_dims for f in log_factors]) + "->"
    plates = "".join(
        frozenset().union(
            list(model_trace.plate_to_symbol.values()),
            list(guide_trace.plate_to_symbol.values()),
        )
    )
    (tmc,) = einsum(
        eqn,
        *log_factors,
        plates=plates,
        backend="pyro.ops.einsum.torch_log",
        modulo_total=False
    )
    return tmc


class TraceTMC_ELBO(ELBO):
    """
    A trace-based implementation of Tensor Monte Carlo [1]
    by way of Tensor Variable Elimination [2] that supports:
    - local parallel sampling over any sample site in the model or guide
    - exhaustive enumeration over any sample site in the model or guide

    To take multiple samples, mark the site with
    ``infer={'enumerate': 'parallel', 'num_samples': N}``.
    To configure all sites in a model or guide at once,
    use :func:`~pyro.infer.enum.config_enumerate` .
    To enumerate or sample a sample site in the ``model``,
    mark the site and ensure the site does not appear in the ``guide``.

    This assumes restricted dependency structure on the model and guide:
    variables outside of an :class:`~pyro.plate` can never depend on
    variables inside that :class:`~pyro.plate` .

    References

    [1] `Tensor Monte Carlo: Particle Methods for the GPU Era`,
        Laurence Aitchison (2018)

    [2] `Tensor Variable Elimination for Plated Factor Graphs`,
        Fritz Obermeyer, Eli Bingham, Martin Jankowiak, Justin Chiu, Neeraj Pradhan,
        Alexander Rush, Noah Goodman (2019)
    """

    def _get_trace(self, model, guide, args, kwargs):
        """
        Returns a single trace from the guide, and the model that is run
        against it.
        """
        model_trace, guide_trace = get_importance_trace(
            "flat", self.max_plate_nesting, model, guide, args, kwargs
        )

        if is_validation_enabled():
            check_traceenum_requirements(model_trace, guide_trace)

            has_enumerated_sites = any(
                site["infer"].get("enumerate")
                for trace in (guide_trace, model_trace)
                for name, site in trace.nodes.items()
                if site["type"] == "sample"
            )

            if self.strict_enumeration_warning and not has_enumerated_sites:
                warnings.warn(
                    "Found no sample sites configured for enumeration. "
                    "If you want to enumerate sites, you need to @config_enumerate or set "
                    'infer={"enumerate": "sequential"} or infer={"enumerate": "parallel"}? '
                    "If you do not want to enumerate, consider using Trace_ELBO instead."
                )

        model_trace.compute_score_parts()
        guide_trace.pack_tensors()
        model_trace.pack_tensors(guide_trace.plate_to_symbol)
        return model_trace, guide_trace

    def _get_traces(self, model, guide, args, kwargs):
        """
        Runs the guide and runs the model against the guide with
        the result packaged as a trace generator.
        """
        if self.max_plate_nesting == float("inf"):
            self._guess_max_plate_nesting(model, guide, args, kwargs)
        if self.vectorize_particles:
            guide = self._vectorized_num_particles(guide)
            model = self._vectorized_num_particles(model)

        # Enable parallel enumeration over the vectorized guide and model.
        # The model allocates enumeration dimensions after (to the left of) the guide,
        # accomplished by preserving the _ENUM_ALLOCATOR state after the guide call.
        guide_enum = EnumMessenger(first_available_dim=-1 - self.max_plate_nesting)
        model_enum = EnumMessenger()  # preserve _ENUM_ALLOCATOR state
        guide = guide_enum(guide)
        model = model_enum(model)

        q = queue.LifoQueue()
        guide = poutine.queue(
            guide, q, escape_fn=iter_discrete_escape, extend_fn=iter_discrete_extend
        )
        for i in range(1 if self.vectorize_particles else self.num_particles):
            q.put(poutine.Trace())
            while not q.empty():
                yield self._get_trace(model, guide, args, kwargs)

    def differentiable_loss(self, model, guide, *args, **kwargs):
        """
        :returns: a differentiable estimate of the marginal log-likelihood
        :rtype: torch.Tensor
        :raises ValueError: if the ELBO is not differentiable (e.g. is
            identically zero)

        Computes a differentiable TMC estimate using ``num_particles`` many samples
        (particles).  The result should be infinitely differentiable (as long
        as underlying derivatives have been implemented).
        """
        elbo = 0.0
        for model_trace, guide_trace in self._get_traces(model, guide, args, kwargs):
            elbo_particle = _compute_tmc_estimate(model_trace, guide_trace)
            if is_identically_zero(elbo_particle):
                continue

            elbo = elbo + elbo_particle
        elbo = elbo / self.num_particles

        loss = -elbo
        warn_if_nan(loss, "loss")
        return loss

    def loss(self, model, guide, *args, **kwargs):
        with torch.no_grad():
            loss = self.differentiable_loss(model, guide, *args, **kwargs)
            if is_identically_zero(loss) or not loss.requires_grad:
                return torch_item(loss)
            return loss.item()

    def loss_and_grads(self, model, guide, *args, **kwargs):
        loss = self.differentiable_loss(model, guide, *args, **kwargs)
        if is_identically_zero(loss) or not loss.requires_grad:
            return torch_item(loss)
        loss.backward()
        return loss.item()



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import math

import torch
import torch.nn as nn

from pyro.nn import AutoRegressiveNN

from .. import constraints
from ..torch_transform import TransformModule
from ..util import copy_docs_from


@copy_docs_from(TransformModule)
class Polynomial(TransformModule):
    r"""
    An autoregressive bijective transform as described in Jaini et al. (2019)
    applying following equation element-wise,

        :math:`y_n = c_n + \int^{x_n}_0\sum^K_{k=1}\left(\sum^R_{r=0}a^{(n)}_{r,k}u^r\right)du`

    where :math:`x_n` is the :math:`n`th input, :math:`y_n` is the :math:`n`th
    output, and :math:`c_n\in\mathbb{R}`,
    :math:`\left\{a^{(n)}_{r,k}\in\mathbb{R}\right\}` are learnable parameters
    that are the output of an autoregressive NN inputting
    :math:`x_{\prec n}={x_1,x_2,\ldots,x_{n-1}}`.

    Together with :class:`~pyro.distributions.TransformedDistribution` this provides
    a way to create richer variational approximations.

    Example usage:

    >>> from pyro.nn import AutoRegressiveNN
    >>> input_dim = 10
    >>> count_degree = 4
    >>> count_sum = 3
    >>> base_dist = dist.Normal(torch.zeros(input_dim), torch.ones(input_dim))
    >>> param_dims = [(count_degree + 1)*count_sum]
    >>> arn = AutoRegressiveNN(input_dim, [input_dim*10], param_dims)
    >>> transform = Polynomial(arn, input_dim=input_dim, count_degree=count_degree,
    ... count_sum=count_sum)
    >>> pyro.module("my_transform", transform)  # doctest: +SKIP
    >>> flow_dist = dist.TransformedDistribution(base_dist, [transform])
    >>> flow_dist.sample()  # doctest: +SKIP

    The inverse of this transform does not possess an analytical solution and is
    left unimplemented. However, the inverse is cached when the forward operation is
    called during sampling, and so samples drawn using a polynomial transform can be
    scored.

    :param autoregressive_nn: an autoregressive neural network whose forward call
        returns a tensor of real-valued
        numbers of size (batch_size, (count_degree+1)*count_sum, input_dim)
    :type autoregressive_nn: nn.Module
    :param count_degree: The degree of the polynomial to use for each element-wise
        transformation.
    :type count_degree: int
    :param count_sum: The number of polynomials to sum in each element-wise
        transformation.
    :type count_sum: int

    References:

    [1] Priyank Jaini, Kira A. Shelby, Yaoliang Yu. Sum-of-squares polynomial flow.
    [arXiv:1905.02325]

    """

    domain = constraints.real_vector
    codomain = constraints.real_vector
    bijective = True
    autoregressive = True

    def __init__(self, autoregressive_nn, input_dim, count_degree, count_sum):
        super().__init__(cache_size=1)

        self.arn = autoregressive_nn
        self.input_dim = input_dim
        self.count_degree = count_degree
        self.count_sum = count_sum
        self._cached_logDetJ = None

        self.c = nn.Parameter(torch.Tensor(input_dim))
        self.reset_parameters()

        # Vector of powers of input dimension
        powers = torch.arange(1, count_degree + 2, dtype=torch.get_default_dtype())
        self.register_buffer("powers", powers)

        # Build mask of constants
        mask = self.powers + torch.arange(count_degree + 1).unsqueeze(-1).type_as(
            powers
        )
        power_mask = mask
        mask = mask.reciprocal()

        self.register_buffer("power_mask", power_mask)
        self.register_buffer("mask", mask)

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.c.size(0))
        self.c.data.uniform_(-stdv, stdv)

    def _call(self, x):
        """
        :param x: the input into the bijection
        :type x: torch.Tensor

        Invokes the bijection x=>y; in the prototypical context of a
        :class:`~pyro.distributions.TransformedDistribution` `x` is a sample from
        the base distribution (or the output of a previous transform)
        """
        # Calculate the polynomial coefficients
        # ~ (batch_size, count_sum, count_degree+1, input_dim)
        A = self.arn(x).view(-1, self.count_sum, self.count_degree + 1, self.input_dim)

        # Take cross product of coefficients across degree dim
        # ~ (batch_size, count_sum, count_degree+1, count_degree+1, input_dim)
        coefs = A.unsqueeze(-2) * A.unsqueeze(-3)

        # Calculate output as sum-of-squares polynomial
        x_view = x.view(-1, 1, 1, self.input_dim)
        x_pow_matrix = x_view.pow(self.power_mask.unsqueeze(-1)).unsqueeze(-4)

        # Eq (8) from the paper, expanding the squared term and integrating
        # NOTE: The view_as is necessary because the batch dimensions were collapsed previously
        y = self.c + (coefs * x_pow_matrix * self.mask.unsqueeze(-1)).sum(
            (1, 2, 3)
        ).view_as(x)

        # log(|det(J)|) is calculated by the fundamental theorem of calculus, i.e. remove the constant
        # term and the integral from eq (8) (the equation for this isn't given in the paper)
        x_pow_matrix = x_view.pow(self.power_mask.unsqueeze(-1) - 1).unsqueeze(-4)
        self._cached_logDetJ = torch.log(
            (coefs * x_pow_matrix).sum((1, 2, 3)).view_as(x) + 1e-8
        ).sum(-1)

        return y

    def _inverse(self, y):
        """
        :param y: the output of the bijection
        :type y: torch.Tensor

        Inverts y => x. As noted above, this implementation is incapable of
        inverting arbitrary values `y`; rather it assumes `y` is the result of a
        previously computed application of the bijector to some `x` (which was
        cached on the forward call)
        """

        raise KeyError(
            "Polynomial object expected to find key in intermediates cache but didn't"
        )

    def log_abs_det_jacobian(self, x, y):
        """
        Calculates the elementwise determinant of the log Jacobian
        """
        x_old, y_old = self._cached_x_y
        if x is not x_old or y is not y_old:
            # This call to the parent class Transform will update the cache
            # as well as calling self._call and recalculating y and log_detJ
            self(x)

        return self._cached_logDetJ


def polynomial(input_dim, hidden_dims=None):
    """
    A helper function to create a :class:`~pyro.distributions.transforms.Polynomial`
    object that takes care of constructing an autoregressive network with the
    correct input/output dimensions.

    :param input_dim: Dimension of input variable
    :type input_dim: int
    :param hidden_dims: The desired hidden dimensions of of the autoregressive
        network. Defaults to using [input_dim * 10]

    """

    count_degree = 4
    count_sum = 3
    if hidden_dims is None:
        hidden_dims = [input_dim * 10]
    arn = AutoRegressiveNN(
        input_dim, hidden_dims, param_dims=[(count_degree + 1) * count_sum]
    )
    return Polynomial(
        arn, input_dim=input_dim, count_degree=count_degree, count_sum=count_sum
    )



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import math

import torch

import pyro
import pyro.poutine as poutine
from pyro.infer.elbo import ELBO
from pyro.infer.enum import get_importance_trace
from pyro.infer.util import is_validation_enabled
from pyro.poutine.util import prune_subsample_sites
from pyro.util import check_if_enumerated, check_model_guide_match, warn_if_nan


class ReweightedWakeSleep(ELBO):
    r"""
    An implementation of Reweighted Wake Sleep following reference [1].

    .. note:: Sampling and log_prob evaluation asymptotic complexity:

    1) Using wake-theta and/or wake-phi
        O(`num_particles`) samples from guide,
        O(`num_particles`) `log_prob` evaluations of model and guide

    2) Using sleep-phi
        O(`num_sleep_particles`) samples from model,
        O(`num_sleep_particles`) `log_prob` evaluations of guide

    if 1) and 2) are combined,
        O(`num_particles`) samples from the guide,
        O(`num_sleep_particles`) from the model,
        O(`num_particles` + `num_sleep_particles`) `log_prob` evaluations of the guide, and
        O(`num_particles`) evaluations of the model

    .. note:: This is particularly useful for models with stochastic branching,
        as described in [2].

    .. note:: This returns _two_ losses, one each for (a) the model parameters (`theta`), computed using the
        `iwae` objective, and (b) the guide parameters (`phi`), computed using (a combination of) the `csis`
        objective and a self-normalized importance-sampled version of the `csis` objective.

    .. note:: In order to enable computing the sleep-phi terms, the guide program must have its observations
        explicitly passed in through the keyworded argument `observations`. Where the value of the observations
        is unknown during definition, such as for amortized variational inference, it may be given a default
        argument as `observations=None`, and the correct value supplied during learning through
        `svi.step(observations=...)`.

    .. warning:: Mini-batch training is not supported yet.

    :param int num_particles: The number of particles/samples used to form the objective
        (gradient) estimator. Default is 2.
    :param insomnia: The scaling between the wake-phi and sleep-phi terms. Default is 1.0 [wake-phi]
    :param bool model_has_params: Indicate if model has learnable params. Useful in avoiding extra
        computation when running in pure sleep mode [csis]. Default is True.
    :param int num_sleep_particles: The number of particles used to form the sleep-phi estimator.
        Matches `num_particles` by default.
    :param bool vectorize_particles: Whether the traces should be vectorised
        across `num_particles`. Default is True.
    :param int max_plate_nesting: Bound on max number of nested
        :func:`pyro.plate` contexts. Default is infinity.
    :param bool strict_enumeration_warning: Whether to warn about possible
        misuse of enumeration, i.e. that
        :class:`~pyro.infer.traceenum_elbo.TraceEnum_ELBO` is used iff there
        are enumerated sample sites.

    References:

    [1] `Reweighted Wake-Sleep`,
        Jörg Bornschein, Yoshua Bengio

    [2] `Revisiting Reweighted Wake-Sleep for Models with Stochastic Control Flow`,
        Tuan Anh Le, Adam R. Kosiorek, N. Siddharth, Yee Whye Teh, Frank Wood
    """

    def __init__(
        self,
        num_particles=2,
        insomnia=1.0,
        model_has_params=True,
        num_sleep_particles=None,
        vectorize_particles=True,
        max_plate_nesting=float("inf"),
        strict_enumeration_warning=True,
    ):
        # force K > 1 otherwise SNIS not possible
        assert (
            num_particles > 1
        ), "Reweighted Wake Sleep needs to be run with more than one particle"

        super().__init__(
            num_particles=num_particles,
            max_plate_nesting=max_plate_nesting,
            vectorize_particles=vectorize_particles,
            strict_enumeration_warning=strict_enumeration_warning,
        )
        self.insomnia = insomnia
        self.model_has_params = model_has_params
        self.num_sleep_particles = (
            num_particles if num_sleep_particles is None else num_sleep_particles
        )

        assert insomnia >= 0 and insomnia <= 1, "insomnia should be in [0, 1]"

    def _get_trace(self, model, guide, args, kwargs):
        """
        Returns a single trace from the guide, and the model that is run against it.
        """
        model_trace, guide_trace = get_importance_trace(
            "flat", self.max_plate_nesting, model, guide, args, kwargs, detach=True
        )
        if is_validation_enabled():
            check_if_enumerated(guide_trace)
        return model_trace, guide_trace

    def _loss(self, model, guide, args, kwargs):
        """
        :returns: returns model loss and guide loss
        :rtype: float, float

        Computes the re-weighted wake-sleep estimators for the model (wake-theta) and the
          guide (insomnia * wake-phi + (1 - insomnia) * sleep-phi).
        Performs backward as appropriate on both, over the specified number of particles.
        """

        wake_theta_loss = torch.tensor(100.0)
        if self.model_has_params or self.insomnia > 0.0:
            # compute quantities for wake theta and wake phi
            log_joints = []
            log_qs = []

            for model_trace, guide_trace in self._get_traces(
                model, guide, args, kwargs
            ):
                log_joint = 0.0
                log_q = 0.0

                for _, site in model_trace.nodes.items():
                    if site["type"] == "sample":
                        if self.vectorize_particles:
                            log_p_site = (
                                site["log_prob"].reshape(self.num_particles, -1).sum(-1)
                            )
                        else:
                            log_p_site = site["log_prob_sum"]
                        log_joint = log_joint + log_p_site

                for _, site in guide_trace.nodes.items():
                    if site["type"] == "sample":
                        if self.vectorize_particles:
                            log_q_site = (
                                site["log_prob"].reshape(self.num_particles, -1).sum(-1)
                            )
                        else:
                            log_q_site = site["log_prob_sum"]
                        log_q = log_q + log_q_site

                log_joints.append(log_joint)
                log_qs.append(log_q)

            log_joints = (
                log_joints[0] if self.vectorize_particles else torch.stack(log_joints)
            )
            log_qs = log_qs[0] if self.vectorize_particles else torch.stack(log_qs)
            log_weights = log_joints - log_qs.detach()

            # compute wake theta loss
            log_sum_weight = torch.logsumexp(log_weights, dim=0)
            wake_theta_loss = -(log_sum_weight - math.log(self.num_particles)).sum()
            warn_if_nan(wake_theta_loss, "wake theta loss")

        if self.insomnia > 0:
            # compute wake phi loss
            normalised_weights = (log_weights - log_sum_weight).exp().detach()
            wake_phi_loss = -(normalised_weights * log_qs).sum()
            warn_if_nan(wake_phi_loss, "wake phi loss")

        if self.insomnia < 1:
            # compute sleep phi loss
            _model = pyro.poutine.uncondition(model)
            _guide = guide
            _log_q = 0.0

            if self.vectorize_particles:
                if self.max_plate_nesting == float("inf"):
                    self._guess_max_plate_nesting(_model, _guide, args, kwargs)
                _model = self._vectorized_num_sleep_particles(_model)
                _guide = self._vectorized_num_sleep_particles(guide)

            for _ in range(1 if self.vectorize_particles else self.num_sleep_particles):
                _model_trace = poutine.trace(_model).get_trace(*args, **kwargs)
                _model_trace.detach_()
                _guide_trace = self._get_matched_trace(
                    _model_trace, _guide, args, kwargs
                )
                _log_q += _guide_trace.log_prob_sum()

            sleep_phi_loss = -_log_q / self.num_sleep_particles
            warn_if_nan(sleep_phi_loss, "sleep phi loss")

        # compute phi loss
        phi_loss = (
            sleep_phi_loss
            if self.insomnia == 0
            else (
                wake_phi_loss
                if self.insomnia == 1
                else self.insomnia * wake_phi_loss
                + (1.0 - self.insomnia) * sleep_phi_loss
            )
        )

        return wake_theta_loss, phi_loss

    def loss(self, model, guide, *args, **kwargs):
        """
        :returns: returns model loss and guide loss
        :rtype: float, float

        Computes the re-weighted wake-sleep estimators for the model (wake-theta) and the
          guide (insomnia * wake-phi + (1 - insomnia) * sleep-phi).
        """
        with torch.no_grad():
            wake_theta_loss, phi_loss = self._loss(model, guide, args, kwargs)

        return wake_theta_loss, phi_loss

    def loss_and_grads(self, model, guide, *args, **kwargs):
        """
        :returns: returns model loss and guide loss
        :rtype: float

        Computes the RWS estimators for the model (wake-theta) and the guide (wake-phi).
        Performs backward as appropriate on both, using num_particle many samples/particles.
        """
        wake_theta_loss, phi_loss = self._loss(model, guide, args, kwargs)
        # convenience addition to ensure easier gradients without requiring `retain_graph=True`
        (wake_theta_loss + phi_loss).backward()

        return wake_theta_loss.detach().item(), phi_loss.detach().item()

    def _vectorized_num_sleep_particles(self, fn):
        """
        Copy of `_vectorised_num_particles` that uses `num_sleep_particles`.
        """

        def wrapped_fn(*args, **kwargs):
            if self.num_sleep_particles == 1:
                return fn(*args, **kwargs)
            with pyro.plate(
                "num_sleep_particles_vectorized",
                self.num_sleep_particles,
                dim=-self.max_plate_nesting,
            ):
                return fn(*args, **kwargs)

        return wrapped_fn

    @staticmethod
    def _get_matched_trace(model_trace, guide, args, kwargs):
        kwargs["observations"] = {}
        for node in model_trace.stochastic_nodes + model_trace.observation_nodes:
            if "was_observed" in model_trace.nodes[node]["infer"]:
                model_trace.nodes[node]["is_observed"] = True
                kwargs["observations"][node] = model_trace.nodes[node]["value"]

        guide_trace = poutine.trace(poutine.replay(guide, model_trace)).get_trace(
            *args, **kwargs
        )
        check_model_guide_match(model_trace, guide_trace)
        guide_trace = prune_subsample_sites(guide_trace)

        return guide_trace



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

"""
Example usage::

    # Simple getting and setting.
    print(pyro.settings.get())  # print all settings
    print(pyro.settings.get("cholesky_relative_jitter"))  # print one
    pyro.settings.set(cholesky_relative_jitter=0.5)  # set one
    pyro.settings.set(**my_settings)  # set many

    # Use as a contextmanager.
    with pyro.settings.context(cholesky_relative_jitter=0.5):
        my_function()

    # Use as a decorator.
    fn = pyro.settings.context(cholesky_relative_jitter=0.5)(my_function)
    fn()

    # Register a new setting.
    pyro.settings.register(
        "binomial_approx_sample_thresh",  # alias
        "pyro.distributions.torch",       # module
        "Binomial.approx_sample_thresh",  # deep name
    )

    # Register a new setting on a user-provided validator.
    @pyro.settings.register(
        "binomial_approx_sample_thresh",  # alias
        "pyro.distributions.torch",       # module
        "Binomial.approx_sample_thresh",  # deep name
    )
    def validate_thresh(thresh):  # called each time setting is set
        assert isinstance(thresh, float)
        assert thresh > 0

Default Settings
----------------

{defaults}

Settings Interface
------------------
"""

# This library must have no dependencies on other pyro modules.
import functools
from contextlib import contextmanager
from importlib import import_module
from typing import Any, Callable, Dict, Iterator, Optional, Tuple

# Docs are updated by register().
_doc_template = __doc__

# Global registry mapping alias:str to (modulename, deepname, validator)
# triples where deepname may have dots to indicate e.g. class variables.
_REGISTRY: Dict[str, Tuple[str, str, Optional[Callable]]] = {}


def get(alias: Optional[str] = None) -> Any:
    """
    Gets one or all global settings.

    :param str alias: The name of a registered setting.
    :returns: The currently set value.
    """
    if alias is None:
        # Return dict of all settings.
        return {alias: get(alias) for alias in sorted(_REGISTRY)}
    # Get a single setting.
    module, deepname, validator = _REGISTRY[alias]
    value = import_module(module)
    for name in deepname.split("."):
        value = getattr(value, name)
    return value


def set(**kwargs) -> None:
    r"""
    Sets one or more settings.

    :param \*\*kwargs: alias=value pairs.
    """
    for alias, value in kwargs.items():
        module, deepname, validator = _REGISTRY[alias]
        if validator is not None:
            validator(value)
        destin = import_module(module)
        names = deepname.split(".")
        for name in names[:-1]:
            destin = getattr(destin, name)
        setattr(destin, names[-1], value)


@contextmanager
def context(**kwargs) -> Iterator[None]:
    r"""
    Context manager to temporarily override one or more settings. This also
    works as a decorator.

    :param \*\*kwargs: alias=value pairs.
    """
    old = {alias: get(alias) for alias in kwargs}
    try:
        set(**kwargs)
        yield
    finally:
        set(**old)


def register(
    alias: str,
    modulename: str,
    deepname: str,
    validator: Optional[Callable] = None,
) -> Callable:
    """
    Register a global settings.

    This should be declared in the module where the setting is defined.

    This can be used either as a declaration::

        settings.register("my_setting", __name__, "MY_SETTING")

    or as a decorator on a user-defined validator function::

        @settings.register("my_setting", __name__, "MY_SETTING")
        def _validate_my_setting(value):
            assert isinstance(value, float)
            assert 0 < value

    :param str alias: A valid python identifier serving as a settings alias.
        Lower snake case preferred, e.g. ``my_setting``.
    :param str modulename: The module name where the setting is declared,
        typically ``__name__``.
    :param str deepname: A ``.``-separated string of names. E.g. for a module
        constant, use ``MY_CONSTANT``. For a class attributue, use
        ``MyClass.my_attribute``.
    :param callable validator: Optional validator that inputs a value,
        possibly raises validation errors, and returns None.
    """
    global __doc__
    assert isinstance(alias, str)
    assert alias.isidentifier()
    assert isinstance(modulename, str)
    assert isinstance(deepname, str)
    _REGISTRY[alias] = modulename, deepname, validator

    # Add default value to module docstring.
    __doc__ = _doc_template.format(
        defaults="\n".join(f"- {a} = {get(a)}" for a in sorted(_REGISTRY))
    )

    # Support use as a decorator on an optional user-provided validator.
    if validator is None:
        # Return a decorator, but its fine if user discards this.
        return functools.partial(register, alias, modulename, deepname)
    else:
        # Test current value passes validation.
        validator(get(alias))
        return validator



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

from contextlib import ExitStack

from torch.distributions import biject_to
from torch.distributions.transforms import ComposeTransform

import pyro
import pyro.distributions as dist
from pyro.poutine.plate_messenger import block_plate

from .reparam import Reparam


class UnitJacobianReparam(Reparam):
    """
    Reparameterizer for :class:`~torch.distributions.transforms.Transform`
    objects whose Jacobian determinant is one.

    :param transform: A transform whose Jacobian has determinant 1.
    :type transform: ~torch.distributions.transforms.Transform
    :param str suffix: A suffix to append to the transformed site.
    :param bool experimental_allow_batch: EXPERIMENTAL allow coupling across a
        batch dimension. The targeted batch dimension and all batch dimensions
        to the right will be converted to event dimensions. Defaults to False.
    """

    def __init__(
        self, transform, suffix="transformed", *, experimental_allow_batch=False
    ):
        self.transform = transform.with_cache()
        self.suffix = suffix
        self.experimental_allow_batch = experimental_allow_batch

    def apply(self, msg):
        name = msg["name"]
        fn = msg["fn"]
        value = msg["value"]
        is_observed = msg["is_observed"]

        event_dim = fn.event_dim
        transform = self.transform
        with ExitStack() as stack:
            shift = max(0, transform.event_dim - event_dim)
            if shift:
                if not self.experimental_allow_batch:
                    raise ValueError(
                        "Cannot transform along batch dimension; try either"
                        "converting a batch dimension to an event dimension, or "
                        "setting experimental_allow_batch=True."
                    )

                # Reshape and mute plates using block_plate.
                from pyro.contrib.forecast.util import (
                    reshape_batch,
                    reshape_transform_batch,
                )

                old_shape = fn.batch_shape
                new_shape = old_shape[:-shift] + (1,) * shift + old_shape[-shift:]
                fn = reshape_batch(fn, new_shape).to_event(shift)
                transform = reshape_transform_batch(
                    transform, old_shape + fn.event_shape, new_shape + fn.event_shape
                )
                if value is not None:
                    value = value.reshape(
                        value.shape[: -shift - event_dim]
                        + (1,) * shift
                        + value.shape[-shift - event_dim :]
                    )
                for dim in range(-shift, 0):
                    stack.enter_context(block_plate(dim=dim, strict=False))

            # Differentiably invert transform.
            transform = ComposeTransform(
                [biject_to(fn.support).inv.with_cache(), self.transform]
            )
            value_trans = None
            if value is not None:
                value_trans = transform(value)

            # Draw noise from the base distribution.
            value_trans = pyro.sample(
                f"{name}_{self.suffix}",
                dist.TransformedDistribution(fn, transform),
                obs=value_trans,
                infer={"is_observed": is_observed},
            )

        # Differentiably transform. This should be free due to transform cache.
        if value is None:
            value = transform.inv(value_trans)
        if shift:
            value = value.reshape(
                value.shape[: -2 * shift - event_dim]
                + value.shape[-shift - event_dim :]
            )

        # Simulate a pyro.deterministic() site.
        new_fn = dist.Delta(value, event_dim=event_dim)
        return {"fn": new_fn, "value": value, "is_observed": True}



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
# Copyright (c) 2015, Gamelan Labs, Inc.
# Copyright (c) 2016, Google, Inc.
# Copyright (c) 2019, Gamalon, Inc.
# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0 AND BSD-3-Clause

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Goodness of Fit Testing
-----------------------

This module implements goodness of fit tests for checking agreement between
distributions' ``.sample()`` and ``.log_prob()`` methods. The main functions
return a goodness of fit p-value ``gof`` which for good data should be
``Uniform(0,1)`` distributed and for bad data should be close to zero. To use
this returned number in tests, set a global variable ``TEST_FAILURE_RATE`` to
something smaller than 1 / number of tests in your suite, then in each test
assert ``gof > TEST_FAILURE_RATE``. For example::

    TEST_FAILURE_RATE = 1 / 20  # For 1 in 20 chance of spurious failure.

    def test_my_distribution():
        d = MyDistribution()
        samples = d.sample([10000])
        probs = d.log_prob(samples).exp()
        gof = auto_goodness_of_fit(samples, probs)
        assert gof > TEST_FAILURE_RATE

This module is a port of the
`goftests <https://github.com/posterior/goftests>`_ library.
"""

import math
import warnings

import torch

from .special import chi2sf

HISTOGRAM_WIDTH = 60


class InvalidTest(ValueError):
    pass


def print_histogram(probs, counts):
    max_count = int(max(counts))
    print("{: >8} {: >8}".format("Prob", "Count"))
    for prob, count in sorted(zip(probs, counts), reverse=True):
        width = int(round(HISTOGRAM_WIDTH * int(count) / max_count))
        print("{: >8.3f} {: >8d} {}".format(prob, count, "-" * width))


@torch.no_grad()
def multinomial_goodness_of_fit(
    probs,
    counts,
    *,
    total_count=None,
    plot=False,
):
    """
    Pearson's chi^2 test, on possibly truncated data.
    https://en.wikipedia.org/wiki/Pearson%27s_chi-squared_test

    :param torch.Tensor probs: Vector of probabilities.
    :param torch.Tensor counts: Vector of counts.
    :param int total_count: Optional total count in case data is truncated,
        otherwise None.
    :param bool plot: Whether to print a histogram. Defaults to False.
    :returns: p-value of truncated multinomial sample.
    :rtype: float
    """
    assert probs.dim() == 1
    assert probs.shape == counts.shape
    if total_count is None:
        truncated = False
        total_count = int(counts.sum())
    else:
        truncated = True
        assert total_count >= counts.sum()
    if plot:
        print_histogram(probs, counts)

    chi_squared = 0
    dof = 0
    for p, c in zip(probs.tolist(), counts.tolist()):
        if abs(p - 1) < 1e-8:
            return 1 if c == total_count else 0
        assert p < 1, f"bad probability: {p:g}"
        if p > 0:
            mean = total_count * p
            variance = total_count * p * (1 - p)
            if not (variance > 1):
                raise InvalidTest("Goodness of fit is inaccurate; use more samples")
            chi_squared += (c - mean) ** 2 / variance
            dof += 1
        else:
            warnings.warn("Zero probability in goodness-of-fit test")
            if c > 0:
                return math.inf

    if not truncated:
        dof -= 1

    survival = chi2sf(chi_squared, dof)
    return survival


@torch.no_grad()
def unif01_goodness_of_fit(samples, *, plot=False):
    """
    Bin uniformly distributed samples and apply Pearson's chi^2 test.

    :param torch.Tensor samples: A vector of real-valued samples from a
        candidate distribution that should be Uniform(0, 1)-distributed.
    :param bool plot: Whether to print a histogram. Defaults to False.
    :returns: Goodness of fit, as a p-value.
    :rtype: float
    """
    assert samples.min() >= 0
    assert samples.max() <= 1
    bin_count = int(round(len(samples) ** 0.333))
    if bin_count < 7:
        raise InvalidTest("imprecise test, use more samples")
    probs = torch.ones(bin_count) / bin_count
    binned = samples.mul(bin_count).long().clamp(min=0, max=bin_count - 1)
    counts = torch.zeros(bin_count)
    counts.scatter_add_(0, binned, torch.ones(binned.shape))
    return multinomial_goodness_of_fit(probs, counts, plot=plot)


@torch.no_grad()
def exp_goodness_of_fit(samples, plot=False):
    """
    Transform exponentially distribued samples to Uniform(0,1) distribution and
    assess goodness of fit via binned Pearson's chi^2 test.

    :param torch.Tensor samples: A vector of real-valued samples from a
        candidate distribution that should be Exponential(1)-distributed.
    :param bool plot: Whether to print a histogram. Defaults to False.
    :returns: Goodness of fit, as a p-value.
    :rtype: float
    """
    unif01_samples = samples.neg().exp()
    return unif01_goodness_of_fit(unif01_samples, plot=plot)


@torch.no_grad()
def density_goodness_of_fit(samples, probs, plot=False):
    """
    Transform arbitrary continuous samples to Uniform(0,1) distribution and
    assess goodness of fit via binned Pearson's chi^2 test.

    :param torch.Tensor samples: A vector list of real-valued samples from a
        distribution.
    :param torch.Tensor probs: A vector of probability densities evaluated at
        those samples.
    :param bool plot: Whether to print a histogram. Defaults to False.
    :returns: Goodness of fit, as a p-value.
    :rtype: float
    """
    assert samples.shape == probs.shape
    if len(samples) <= 100:
        raise InvalidTest("imprecision; use more samples")

    samples, index = samples.sort(0)
    probs = probs[index]
    gaps = samples[1:] - samples[:-1]

    sparsity = 1 / probs
    sparsity = 0.5 * (sparsity[1:] + sparsity[:-1])
    density = len(samples) / sparsity

    exp_samples = density * gaps
    return exp_goodness_of_fit(exp_samples, plot=plot)


def volume_of_sphere(dim, radius):
    return radius**dim * math.pi ** (0.5 * dim) / math.gamma(0.5 * dim + 1)


def get_nearest_neighbor_distances(samples):
    try:
        # This version scales as O(N log(N)).
        from scipy.spatial import cKDTree

        samples = samples.cpu().numpy()
        distances, indices = cKDTree(samples).query(samples, k=2)
        return torch.from_numpy(distances[:, 1])
    except ImportError:
        # This version scales as O(N^2).
        distances, indices = torch.cdist(samples, samples).kthvalue(k=2)
        return distances


@torch.no_grad()
def vector_density_goodness_of_fit(samples, probs, *, dim=None, plot=False):
    """
    Transform arbitrary multivariate continuous samples to Univariate(0,1)
    distribution via nearest neighbor distribution [1,2,3] and assess goodness
    of fit via binned Pearson's chi^2 test.

    [1] Peter J. Bickel and Leo Breiman (1983)
        "Sums of Functions of Nearest Neighbor Distances, Moment Bounds, Limit
        Theorems and a Goodness of Fit Test"
        https://projecteuclid.org/download/pdf_1/euclid.aop/1176993668
    [2] Mike Williams (2010)
        "How good are your fits? Unbinned multivariate goodness-of-fit tests in
        high energy physics."
        https://arxiv.org/abs/1006.3019
    [3] Nearest Neighbour Distribution
        https://en.wikipedia.org/wiki/Nearest_neighbour_distribution

    :param torch.Tensor samples: A tensor of real-vector-valued samples from a
        distribution.
    :param torch.Tensor probs: A vector of probability densities evaluated at
        those samples.
    :param int dim: Optional dimension of the submanifold on which data lie.
        Defaults to ``samples.shape[-1]``.
    :param bool plot: Whether to print a histogram. Defaults to False.
    :returns: Goodness of fit, as a p-value.
    :rtype: float
    """
    assert samples.shape and len(samples)
    assert probs.shape == samples.shape[:1]
    if dim is None:
        dim = samples.shape[-1]
    assert dim
    if len(samples) <= 1000 * dim:
        raise InvalidTest("imprecision; use more samples")
    radii = get_nearest_neighbor_distances(samples)
    density = len(samples) * probs
    volume = volume_of_sphere(dim, radii)
    exp_samples = density * volume
    return exp_goodness_of_fit(exp_samples, plot=plot)


@torch.no_grad()
def auto_goodness_of_fit(samples, probs, *, dim=None, plot=False):
    """
    Dispatch on sample dimension and delegate to either
    :func:`density_goodness_of_fit` or :func:`vector_density_goodness_of_fit`.

    :param torch.Tensor samples: A tensor of samples stacked on their leftmost
        dimension.
    :param torch.Tensor probs: A vector of probabilities evaluated at those
        samples.
    :param int dim: Optional manifold dimension, defaults to
        ``samples.shape[1:].numel()``.
    :param bool plot: Whether to print a histogram. Defaults to False.
    """
    assert samples.shape and samples.shape[0]
    assert probs.shape == samples.shape[:1]

    samples = samples.reshape(samples.shape[0], -1)
    ambient_dim = samples.shape[1:].numel()
    if dim is None:
        dim = ambient_dim

    if ambient_dim == 0:
        return 1.0
    if ambient_dim == 1:
        samples = samples.reshape(-1)
        return density_goodness_of_fit(samples, probs, plot=plot)
    return vector_density_goodness_of_fit(samples, probs, dim=dim, plot=plot)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import torch

from .torch import MultivariateNormal, Normal


class NanMaskedNormal(Normal):
    """
    Wrapper around :class:`~pyro.distributions.Normal` to allow partially
    observed data as specified by NAN elements in :meth:`log_prob`; the
    ``log_prob`` of these elements will be zero. This is useful for likelihoods
    with missing data.

    Example::

        from math import nan
        data = torch.tensor([0.5, 0.1, nan, 0.9])
        with pyro.plate("data", len(data)):
            pyro.sample("obs", NanMaskedNormal(0, 1), obs=data)
    """

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        ok = value.isfinite()
        if ok.all():
            return super().log_prob(value)

        # Broadcast all tensors.
        value, ok, loc, scale = torch.broadcast_tensors(value, ok, self.loc, self.scale)
        result = value.new_zeros(value.shape)

        # Evaluate ok elements.
        if ok.any():
            marginal = Normal(loc[ok], scale[ok], validate_args=False)
            result[ok] = marginal.log_prob(value[ok])
        return result


class NanMaskedMultivariateNormal(MultivariateNormal):
    """
    Wrapper around :class:`~pyro.distributions.MultivariateNormal` to allow
    partially observed data as specified by NAN elements in the argument to
    :meth:`log_prob`. The ``log_prob`` of these events will marginalize over
    the NAN elements. This is useful for likelihoods with missing data.

    Example::

        from math import nan
        data = torch.tensor([
            [0.1, 0.2, 3.4],
            [0.5, 0.1, nan],
            [0.6, nan, nan],
            [nan, 0.5, nan],
            [nan, nan, nan],
        ])
        with pyro.plate("data", len(data)):
            pyro.sample(
                "obs",
                NanMaskedMultivariateNormal(torch.zeros(3), torch.eye(3)),
                obs=data,
            )
    """

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        ok = value.isfinite()
        if ok.all():
            return super().log_prob(value)

        # Broadcast all tensors. This might waste some computation by eagerly
        # broadcasting, but the optimal implementation is quite complex.
        value, ok, loc = torch.broadcast_tensors(value, ok, self.loc)
        cov = self.covariance_matrix.expand(loc.shape + loc.shape[-1:])

        # Flatten.
        result_shape = value.shape[:-1]
        n = result_shape.numel()
        p = value.shape[-1]
        value = value.reshape(n, p)
        ok = ok.reshape(n, p)
        loc = loc.reshape(n, p)
        cov = cov.reshape(n, p, p)
        result = value.new_zeros(n)

        # Evaluate ok elements.
        for pattern in sorted(set(map(tuple, ok.tolist()))):
            if not any(pattern):
                continue
            # Marginalize out NAN elements.
            col_mask = torch.tensor(pattern)
            row_mask = (ok == col_mask).all(-1)
            ok_value = value[row_mask][:, col_mask]
            ok_loc = loc[row_mask][:, col_mask]
            ok_cov = cov[row_mask][:, col_mask][:, :, col_mask]
            marginal = MultivariateNormal(ok_loc, ok_cov, validate_args=False)
            result[row_mask] = marginal.log_prob(ok_value)

        # Unflatten.
        return result.reshape(result_shape)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import torch
from torch.distributions import constraints

from .torch_distribution import TorchDistribution
from .util import broadcast_shape


class ImproperUniform(TorchDistribution):
    """
    Improper distribution with zero :meth:`log_prob` and undefined
    :meth:`sample`.

    This is useful for transforming a model from generative dag form to factor
    graph form for use in HMC. For example the following are equal in
    distribution::

        # Version 1. a generative dag
        x = pyro.sample("x", Normal(0, 1))
        y = pyro.sample("y", Normal(x, 1))
        z = pyro.sample("z", Normal(y, 1))

        # Version 2. a factor graph
        xyz = pyro.sample("xyz", ImproperUniform(constraints.real, (), (3,)))
        x, y, z = xyz.unbind(-1)
        pyro.sample("x", Normal(0, 1), obs=x)
        pyro.sample("y", Normal(x, 1), obs=y)
        pyro.sample("z", Normal(y, 1), obs=z)

    Note this distribution errors when :meth:`sample` is called. To create a
    similar distribution that instead samples from a specified distribution
    consider using ``.mask(False)`` as in::

        xyz = dist.Normal(0, 1).expand([3]).to_event(1).mask(False)

    :param support: The support of the distribution.
    :type support: ~torch.distributions.constraints.Constraint
    :param torch.Size batch_shape: The batch shape.
    :param torch.Size event_shape: The event shape.
    """

    arg_constraints = {}

    def __init__(self, support, batch_shape, event_shape):
        assert isinstance(support, constraints.Constraint)
        self._support = support
        super().__init__(batch_shape, event_shape)

    @constraints.dependent_property
    def support(self):
        return self._support

    def expand(self, batch_shape, _instance=None):
        batch_shape = torch.Size(batch_shape)
        new = self._get_checked_instance(ImproperUniform, _instance)
        new._support = self._support
        super(ImproperUniform, new).__init__(batch_shape, self.event_shape)
        return new

    def log_prob(self, value):
        batch_shape = value.shape[: value.dim() - self.event_dim]
        batch_shape = broadcast_shape(batch_shape, self.batch_shape)
        return torch.zeros(()).expand(batch_shape)

    def sample(self, sample_shape=torch.Size()):
        raise NotImplementedError("ImproperUniform does not support sampling")



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import math

import torch
from torch.distributions import constraints
from torch.distributions.utils import lazy_property

from pyro.distributions.torch import Chi2
from pyro.distributions.torch_distribution import TorchDistribution
from pyro.distributions.util import broadcast_shape


class MultivariateStudentT(TorchDistribution):
    """
    Creates a multivariate Student's t-distribution parameterized by degree of
    freedom :attr:`df`, mean :attr:`loc` and scale :attr:`scale_tril`.

    :param ~torch.Tensor df: degrees of freedom
    :param ~torch.Tensor loc: mean of the distribution
    :param ~torch.Tensor scale_tril: scale of the distribution, which is
        a lower triangular matrix with positive diagonal entries
    """

    arg_constraints = {
        "df": constraints.positive,
        "loc": constraints.real_vector,
        "scale_tril": constraints.lower_cholesky,
    }
    support = constraints.real_vector
    has_rsample = True

    def __init__(self, df, loc, scale_tril, validate_args=None):
        dim = loc.size(-1)
        assert scale_tril.shape[-2:] == (dim, dim)
        if not isinstance(df, torch.Tensor):
            df = loc.new_tensor(df)
        batch_shape = torch.broadcast_shapes(
            df.shape, loc.shape[:-1], scale_tril.shape[:-2]
        )
        event_shape = torch.Size((dim,))
        self.df = df.expand(batch_shape)
        self.loc = loc.expand(batch_shape + event_shape)
        self._unbroadcasted_scale_tril = scale_tril
        self._chi2 = Chi2(self.df)
        super().__init__(batch_shape, event_shape, validate_args=validate_args)

    @lazy_property
    def scale_tril(self):
        return self._unbroadcasted_scale_tril.expand(
            self._batch_shape + self._event_shape + self._event_shape
        )

    @lazy_property
    def covariance_matrix(self):
        # NB: this is not covariance of this distribution;
        # the actual covariance is df / (df - 2) * covariance_matrix
        return torch.matmul(
            self._unbroadcasted_scale_tril,
            self._unbroadcasted_scale_tril.transpose(-1, -2),
        ).expand(self._batch_shape + self._event_shape + self._event_shape)

    @lazy_property
    def precision_matrix(self):
        identity = torch.eye(
            self.loc.size(-1), device=self.loc.device, dtype=self.loc.dtype
        )
        return torch.cholesky_solve(identity, self._unbroadcasted_scale_tril).expand(
            self._batch_shape + self._event_shape + self._event_shape
        )

    @staticmethod
    def infer_shapes(df, loc, scale_tril):
        event_shape = loc[-1:]
        batch_shape = broadcast_shape(df, loc[:-1], scale_tril[:-2])
        return batch_shape, event_shape

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(MultivariateStudentT, _instance)
        batch_shape = torch.Size(batch_shape)
        loc_shape = batch_shape + self.event_shape
        scale_shape = loc_shape + self.event_shape
        new.df = self.df.expand(batch_shape)
        new.loc = self.loc.expand(loc_shape)
        new._unbroadcasted_scale_tril = self._unbroadcasted_scale_tril
        if "scale_tril" in self.__dict__:
            new.scale_tril = self.scale_tril.expand(scale_shape)
        if "covariance_matrix" in self.__dict__:
            new.covariance_matrix = self.covariance_matrix.expand(scale_shape)
        if "precision_matrix" in self.__dict__:
            new.precision_matrix = self.precision_matrix.expand(scale_shape)
        new._chi2 = self._chi2.expand(batch_shape)
        super(MultivariateStudentT, new).__init__(
            batch_shape, self.event_shape, validate_args=False
        )
        new._validate_args = self._validate_args
        return new

    def rsample(self, sample_shape=torch.Size()):
        shape = self._extended_shape(sample_shape)
        X = torch.empty(shape, dtype=self.df.dtype, device=self.df.device).normal_()
        Z = self._chi2.rsample(sample_shape)
        Y = X * torch.rsqrt(Z / self.df).unsqueeze(-1)
        return self.loc + self.scale_tril.matmul(Y.unsqueeze(-1)).squeeze(-1)

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)
        n = self.loc.size(-1)
        y = torch.linalg.solve_triangular(
            self.scale_tril, (value - self.loc).unsqueeze(-1), upper=False
        ).squeeze(-1)
        Z = (
            self.scale_tril.diagonal(dim1=-2, dim2=-1).log().sum(-1)
            + 0.5 * n * self.df.log()
            + 0.5 * n * math.log(math.pi)
            + torch.lgamma(0.5 * self.df)
            - torch.lgamma(0.5 * (self.df + n))
        )
        return -0.5 * (self.df + n) * torch.log1p(y.pow(2).sum(-1) / self.df) - Z

    @property
    def mean(self):
        m = self.loc.clone()
        m[self.df <= 1, :] = float("nan")
        return m

    @property
    def variance(self):
        m = self.scale_tril.pow(2).sum(-1) * (self.df / (self.df - 2)).unsqueeze(-1)
        m[(self.df <= 2) & (self.df > 1), :] = float("inf")
        m[self.df <= 1, :] = float("nan")
        return m



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import math
import warnings

import torch
from torch.distributions.transforms import CorrCholeskyTransform, Transform

from .. import constraints


class CorrLCholeskyTransform(CorrCholeskyTransform):  # DEPRECATED
    def __init__(self, cache_size=0):
        warnings.warn(
            "class CorrLCholeskyTransform is deprecated in favor of CorrCholeskyTransform.",
            FutureWarning,
        )
        super().__init__(cache_size=cache_size)


class CholeskyTransform(Transform):
    r"""
    Transform via the mapping :math:`y = safe_cholesky(x)`, where `x` is a
    positive definite matrix.
    """

    bijective = True
    domain = constraints.positive_definite
    codomain = constraints.lower_cholesky

    def __eq__(self, other):
        return isinstance(other, CholeskyTransform)

    def _call(self, x):
        return torch.linalg.cholesky(x)

    def _inverse(self, y):
        return torch.matmul(y, torch.transpose(y, -2, -1))

    def log_abs_det_jacobian(self, x, y):
        # Ref: http://web.mit.edu/18.325/www/handouts/handout2.pdf page 13
        n = x.shape[-1]
        order = torch.arange(n, 0, -1, dtype=x.dtype, device=x.device)
        return -n * math.log(2) - (
            order * torch.diagonal(y, dim1=-2, dim2=-1).log()
        ).sum(-1)


class CorrMatrixCholeskyTransform(CholeskyTransform):
    r"""
    Transform via the mapping :math:`y = safe_cholesky(x)`, where `x` is a
    correlation matrix.
    """

    bijective = True
    domain = constraints.corr_matrix
    # TODO: change corr_cholesky_constraint to corr_cholesky when the latter is availabler
    codomain = constraints.corr_cholesky_constraint

    def __eq__(self, other):
        return isinstance(other, CorrMatrixCholeskyTransform)

    def log_abs_det_jacobian(self, x, y):
        # NB: see derivation in LKJCholesky implementation
        n = x.shape[-1]
        order = torch.arange(n - 1, -1, -1, dtype=x.dtype, device=x.device)
        return -(order * torch.diagonal(y, dim1=-2, dim2=-1).log()).sum(-1)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

"""
This module offers a modified interface for MCMC inference with the following objectives:
  - making MCMC independent of Pyro specific trace data structure, to facilitate
    integration with other PyTorch based libraries.
  - bringing the interface closer to that of NumPyro to make it easier to write
    code that works with different backends.
  - minimal memory consumption with multiprocessing and CUDA.
"""
import copy
import json
import logging
import queue
import signal
import threading
import warnings
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Dict

import torch
import torch.multiprocessing as mp

import pyro
import pyro.poutine as poutine
from pyro.infer.mcmc.hmc import HMC
from pyro.infer.mcmc.logger import (
    DIAGNOSTIC_MSG,
    ProgressBar,
    TqdmHandler,
    initialize_logger,
)
from pyro.infer.mcmc.nuts import NUTS
from pyro.infer.mcmc.util import (
    diagnostics,
    diagnostics_from_stats,
    print_summary,
    select_samples,
)
from pyro.ops.streaming import CountMeanVarianceStats, StatsOfDict, StreamingStats
from pyro.util import optional

MAX_SEED = 2**32 - 1


def logger_thread(
    log_queue, warmup_steps, num_samples, num_chains, disable_progbar=False
):
    """
    Logging thread that asynchronously consumes logging events from `log_queue`,
    and handles them appropriately.
    """
    progress_bars = ProgressBar(
        warmup_steps, num_samples, disable=disable_progbar, num_bars=num_chains
    )
    logger = logging.getLogger(__name__)
    logger.propagate = False
    logger.addHandler(TqdmHandler())
    num_samples = [0] * num_chains
    try:
        while True:
            try:
                record = log_queue.get(timeout=1)
            except queue.Empty:
                continue
            if record is None:
                break
            metadata, msg = record.getMessage().split("]", 1)
            _, msg_type, logger_id = metadata[1:].split()
            if msg_type == DIAGNOSTIC_MSG:
                pbar_pos = int(logger_id.split(":")[-1])
                num_samples[pbar_pos] += 1
                if num_samples[pbar_pos] == warmup_steps:
                    progress_bars.set_description(
                        "Sample [{}]".format(pbar_pos + 1), pos=pbar_pos
                    )
                diagnostics = json.loads(msg, object_pairs_hook=OrderedDict)
                progress_bars.set_postfix(diagnostics, pos=pbar_pos, refresh=False)
                progress_bars.update(pos=pbar_pos)
            else:
                logger.handle(record)
    finally:
        progress_bars.close()


class _Worker:
    def __init__(
        self,
        chain_id,
        result_queue,
        log_queue,
        event,
        kernel,
        num_samples,
        warmup_steps,
        initial_params=None,
        hook=None,
    ):
        self.chain_id = chain_id
        self.kernel = kernel
        if initial_params is not None:
            self.kernel.initial_params = initial_params
        self.num_samples = num_samples
        self.warmup_steps = warmup_steps
        self.rng_seed = (torch.initial_seed() + chain_id) % MAX_SEED
        self.log_queue = log_queue
        self.result_queue = result_queue
        self.default_dtype = torch.Tensor().dtype
        self.default_device = torch.Tensor().device
        self.hook = hook
        self.event = event

    def run(self, *args, **kwargs):
        pyro.set_rng_seed(self.rng_seed)
        torch.set_default_dtype(self.default_dtype)
        torch.set_default_device(self.default_device)
        kwargs = kwargs
        logger = logging.getLogger("pyro.infer.mcmc")
        logger_id = "CHAIN:{}".format(self.chain_id)
        log_queue = self.log_queue
        logger = initialize_logger(logger, logger_id, None, log_queue)
        logging_hook = _add_logging_hook(logger, None, self.hook)

        try:
            for sample in _gen_samples(
                self.kernel,
                self.warmup_steps,
                self.num_samples,
                logging_hook,
                None,
                *args,
                **kwargs
            ):
                self.result_queue.put_nowait((self.chain_id, sample))
                self.event.wait()
                self.event.clear()
            self.result_queue.put_nowait((self.chain_id, None))
        except Exception as e:
            logger.exception(e)
            self.result_queue.put_nowait((self.chain_id, e))


def _gen_samples(kernel, warmup_steps, num_samples, hook, chain_id, *args, **kwargs):
    kernel.setup(warmup_steps, *args, **kwargs)
    params = kernel.initial_params
    save_params = getattr(kernel, "save_params", sorted(params))
    # yield structure (key, value.shape) of params
    yield {name: params[name].shape for name in save_params}
    for i in range(warmup_steps):
        params = kernel.sample(params)
        hook(
            kernel,
            params,
            "Warmup [{}]".format(chain_id) if chain_id is not None else "Warmup",
            i,
        )
    for i in range(num_samples):
        params = kernel.sample(params)
        hook(
            kernel,
            params,
            "Sample [{}]".format(chain_id) if chain_id is not None else "Sample",
            i,
        )
        flat = [params[name].reshape(-1) for name in save_params]
        yield (torch.cat if flat else torch.tensor)(flat)
    yield kernel.diagnostics()
    kernel.cleanup()


def _add_logging_hook(logger, progress_bar=None, hook=None):
    def _add_logging(kernel, params, stage, i):
        diagnostics = json.dumps(kernel.logging())
        logger.info(diagnostics, extra={"msg_type": DIAGNOSTIC_MSG})
        if progress_bar:
            progress_bar.set_description(stage, refresh=False)
        if hook:
            hook(kernel, params, stage, i)

    return _add_logging


class _UnarySampler:
    """
    Single process runner class optimized for the case chains are drawn sequentially.
    """

    def __init__(
        self,
        kernel,
        num_samples,
        warmup_steps,
        num_chains,
        disable_progbar,
        initial_params=None,
        hook=None,
    ):
        self.kernel = kernel
        self.initial_params = initial_params
        self.warmup_steps = warmup_steps
        self.num_samples = num_samples
        self.num_chains = num_chains
        self.logger = None
        self.disable_progbar = disable_progbar
        self.hook = hook
        super().__init__()

    def terminate(self, *args, **kwargs):
        pass

    def run(self, *args, **kwargs):
        logger = logging.getLogger("pyro.infer.mcmc")
        for i in range(self.num_chains):
            if self.initial_params is not None:
                initial_params = {k: v[i] for k, v in self.initial_params.items()}
                self.kernel.initial_params = initial_params

            progress_bar = ProgressBar(
                self.warmup_steps, self.num_samples, disable=self.disable_progbar
            )
            logger = initialize_logger(logger, "", progress_bar)
            hook_w_logging = _add_logging_hook(logger, progress_bar, self.hook)
            for sample in _gen_samples(
                self.kernel,
                self.warmup_steps,
                self.num_samples,
                hook_w_logging,
                i if self.num_chains > 1 else None,
                *args,
                **kwargs
            ):
                yield sample, i  # sample, chain_id
            self.kernel.cleanup()
            progress_bar.close()


class _MultiSampler:
    """
    Parallel runner class for running MCMC chains in parallel. This uses the
    `torch.multiprocessing` module (itself a light wrapper over the python
    `multiprocessing` module) to spin up parallel workers.
    """

    def __init__(
        self,
        kernel,
        num_samples,
        warmup_steps,
        num_chains,
        mp_context,
        disable_progbar,
        initial_params=None,
        hook=None,
    ):
        self.kernel = kernel
        self.warmup_steps = warmup_steps
        self.num_chains = num_chains
        self.hook = hook
        self.workers = []
        self.ctx = mp
        if mp_context:
            self.ctx = mp.get_context(mp_context)
        self.result_queue = self.ctx.Queue()
        self.log_queue = self.ctx.Queue()
        self.logger = initialize_logger(
            logging.getLogger("pyro.infer.mcmc"), "MAIN", log_queue=self.log_queue
        )
        self.num_samples = num_samples
        self.initial_params = initial_params
        self.log_thread = threading.Thread(
            target=logger_thread,
            args=(
                self.log_queue,
                self.warmup_steps,
                self.num_samples,
                self.num_chains,
                disable_progbar,
            ),
        )
        self.log_thread.daemon = True
        self.log_thread.start()
        self.events = [self.ctx.Event() for _ in range(num_chains)]

    def init_workers(self, *args, **kwargs):
        self.workers = []
        for i in range(self.num_chains):
            init_params = (
                {k: v[i] for k, v in self.initial_params.items()}
                if self.initial_params is not None
                else None
            )
            worker = _Worker(
                i,
                self.result_queue,
                self.log_queue,
                self.events[i],
                self.kernel,
                self.num_samples,
                self.warmup_steps,
                initial_params=init_params,
                hook=self.hook,
            )
            worker.daemon = True
            self.workers.append(
                self.ctx.Process(
                    name=str(i), target=worker.run, args=args, kwargs=kwargs
                )
            )

    def terminate(self, terminate_workers=False):
        if self.log_thread.is_alive():
            self.log_queue.put_nowait(None)
            self.log_thread.join(timeout=1)
        # Only kill workers if exception is raised. worker processes are daemon
        # processes that will otherwise be terminated with the main process.
        # Note that it is important to not
        if terminate_workers:
            for w in self.workers:
                if w.is_alive():
                    w.terminate()

    def run(self, *args, **kwargs):
        # Ignore sigint in worker processes; they will be shut down
        # when the main process terminates.
        sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
        self.init_workers(*args, **kwargs)
        # restore original handler
        signal.signal(signal.SIGINT, sigint_handler)
        active_workers = self.num_chains
        exc_raised = True
        try:
            for w in self.workers:
                w.start()
            while active_workers:
                try:
                    chain_id, val = self.result_queue.get(timeout=5)
                except queue.Empty:
                    continue
                if isinstance(val, Exception):
                    # Exception trace is already logged by worker.
                    raise val
                if val is not None:
                    yield val, chain_id
                    self.events[chain_id].set()
                else:
                    active_workers -= 1
            exc_raised = False
        finally:
            self.terminate(terminate_workers=exc_raised)


class AbstractMCMC(ABC):
    """
    Base class for MCMC methods.
    """

    def __init__(self, kernel, num_chains, transforms):
        self.kernel = kernel
        self.num_chains = num_chains
        self.transforms = transforms

    @abstractmethod
    def run(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def diagnostics(self):
        raise NotImplementedError

    def _set_transforms(self, *args, **kwargs):
        # Use `kernel.transforms` when available
        if getattr(self.kernel, "transforms", None) is not None:
            self.transforms = self.kernel.transforms
        # Else, get transforms from model (e.g. in multiprocessing).
        elif self.kernel.model:
            warmup_steps = 0
            self.kernel.setup(warmup_steps, *args, **kwargs)
            self.transforms = self.kernel.transforms
        # Assign default value
        else:
            self.transforms = {}

    def _validate_kernel(self, initial_params):
        if (
            isinstance(self.kernel, (HMC, NUTS))
            and self.kernel.potential_fn is not None
        ):
            if initial_params is None:
                raise ValueError(
                    "Must provide valid initial parameters to begin sampling"
                    " when using `potential_fn` in HMC/NUTS kernel."
                )

    def _validate_initial_params(self, initial_params):
        for v in initial_params.values():
            if v.shape[0] != self.num_chains:
                raise ValueError(
                    "The leading dimension of tensors in `initial_params` "
                    "must match the number of chains."
                )


class MCMC(AbstractMCMC):
    """
    Wrapper class for Markov Chain Monte Carlo algorithms. Specific MCMC algorithms
    are TraceKernel instances and need to be supplied as a ``kernel`` argument
    to the constructor.

    .. note:: The case of `num_chains > 1` uses python multiprocessing to
        run parallel chains in multiple processes. This goes with the usual
        caveats around multiprocessing in python, e.g. the model used to
        initialize the ``kernel`` must be serializable via `pickle`, and the
        performance / constraints will be platform dependent (e.g. only
        the "spawn" context is available in Windows). This has also not
        been extensively tested on the Windows platform.

    :param kernel: An instance of the ``TraceKernel`` class, which when
        given an execution trace returns another sample trace from the target
        (posterior) distribution.
    :param int num_samples: The number of samples that need to be generated,
        excluding the samples discarded during the warmup phase.
    :param int warmup_steps: Number of warmup iterations. The samples generated
        during the warmup phase are discarded. If not provided, default is
        is the same as `num_samples`.
    :param int num_chains: Number of MCMC chains to run in parallel. Depending on
        whether `num_chains` is 1 or more than 1, this class internally dispatches
        to either `_UnarySampler` or `_MultiSampler`.
    :param dict initial_params: dict containing initial tensors in unconstrained
        space to initiate the markov chain. The leading dimension's size must match
        that of `num_chains`. If not specified, parameter values will be sampled from
        the prior.
    :param hook_fn: Python callable that takes in `(kernel, samples, stage, i)`
        as arguments. stage is either `sample` or `warmup` and i refers to the
        i'th sample for the given stage. This can be used to implement additional
        logging, or more generally, run arbitrary code per generated sample.
    :param str mp_context: Multiprocessing context to use when `num_chains > 1`.
        Only applicable for Python 3.5 and above. Use `mp_context="spawn"` for
        CUDA.
    :param bool disable_progbar: Disable progress bar and diagnostics update.
    :param bool disable_validation: Disables distribution validation check.
        Defaults to ``True``, disabling validation, since divergent transitions
        will lead to exceptions. Switch to ``False`` to enable validation, or
        to ``None`` to preserve existing global values.
    :param dict transforms: dictionary that specifies a transform for a sample site
        with constrained support to unconstrained space.
    :param List[str] save_params: Optional list of a subset of parameter names to
        save during sampling and diagnostics. This is useful in models with
        large nuisance variables. Defaults to None, saving all params.
    """

    def __init__(
        self,
        kernel,
        num_samples,
        warmup_steps=None,
        initial_params=None,
        num_chains=1,
        hook_fn=None,
        mp_context=None,
        disable_progbar=False,
        disable_validation=True,
        transforms=None,
        save_params=None,
    ):
        super().__init__(kernel, num_chains, transforms)
        self.warmup_steps = (
            num_samples if warmup_steps is None else warmup_steps
        )  # Stan
        self.num_samples = num_samples
        self.disable_validation = disable_validation
        self._samples = None
        self._args = None
        self._kwargs = None
        if save_params is not None:
            kernel.save_params = save_params
        self._validate_kernel(initial_params)
        parallel = False
        if num_chains > 1:
            # check that initial_params is different for each chain
            if initial_params:
                self._validate_initial_params(initial_params)
                # FIXME: probably we want to use "spawn" method by default to avoid the error
                # CUDA initialization error https://github.com/pytorch/pytorch/issues/2517
                # even that we run MCMC in CPU.
                if mp_context is None:
                    # change multiprocessing context to 'spawn' for CUDA tensors.
                    if list(initial_params.values())[0].is_cuda:
                        mp_context = "spawn"

            # verify num_chains is compatible with available CPU.
            available_cpu = max(
                mp.cpu_count() - 1, 1
            )  # reserving 1 for the main process.
            if num_chains <= available_cpu:
                parallel = True
            else:
                warnings.warn(
                    "num_chains={} is more than available_cpu={}. "
                    "Chains will be drawn sequentially.".format(
                        num_chains, available_cpu
                    )
                )
        else:
            if initial_params:
                initial_params = {k: v.unsqueeze(0) for k, v in initial_params.items()}
        self._diagnostics = [None] * num_chains
        if parallel:
            self.sampler = _MultiSampler(
                kernel,
                num_samples,
                self.warmup_steps,
                num_chains,
                mp_context,
                disable_progbar,
                initial_params=initial_params,
                hook=hook_fn,
            )
        else:
            self.sampler = _UnarySampler(
                kernel,
                num_samples,
                self.warmup_steps,
                num_chains,
                disable_progbar,
                initial_params=initial_params,
                hook=hook_fn,
            )

    @poutine.block
    def run(self, *args, **kwargs):
        """
        Run MCMC to generate samples and populate `self._samples`.

        Example usage:

        .. code-block:: python

            def model(data):
                ...

            nuts_kernel = NUTS(model)
            mcmc = MCMC(nuts_kernel, num_samples=500)
            mcmc.run(data)
            samples = mcmc.get_samples()

        :param args: optional arguments taken by
            :meth:`MCMCKernel.setup <pyro.infer.mcmc.mcmc_kernel.MCMCKernel.setup>`.
        :param kwargs: optional keywords arguments taken by
            :meth:`MCMCKernel.setup <pyro.infer.mcmc.mcmc_kernel.MCMCKernel.setup>`.
        """
        self._args, self._kwargs = args, kwargs
        num_samples = [0] * self.num_chains
        z_flat_acc = [[] for _ in range(self.num_chains)]
        with optional(
            pyro.validation_enabled(not self.disable_validation),
            self.disable_validation is not None,
        ):
            # XXX we clone CUDA tensor args to resolve the issue "Invalid device pointer"
            # at https://github.com/pytorch/pytorch/issues/10375
            # This also resolves "RuntimeError: Cowardly refusing to serialize non-leaf tensor which
            # requires_grad", which happens with `jit_compile` under PyTorch 1.7
            args = [arg.detach() if torch.is_tensor(arg) else arg for arg in args]
            for x, chain_id in self.sampler.run(*args, **kwargs):
                if num_samples[chain_id] == 0:
                    num_samples[chain_id] += 1
                    z_structure = x
                elif num_samples[chain_id] == self.num_samples + 1:
                    self._diagnostics[chain_id] = x
                else:
                    num_samples[chain_id] += 1
                    if self.num_chains > 1:
                        x_cloned = x.clone()
                        del x
                    else:
                        x_cloned = x
                    z_flat_acc[chain_id].append(x_cloned)

        z_flat_acc = torch.stack([torch.stack(l) for l in z_flat_acc])

        # unpack latent
        pos = 0
        z_acc = z_structure.copy()
        for k in sorted(z_structure):
            shape = z_structure[k]
            next_pos = pos + shape.numel()
            z_acc[k] = z_flat_acc[:, :, pos:next_pos].reshape(
                (self.num_chains, self.num_samples) + shape
            )
            pos = next_pos
        assert pos == z_flat_acc.shape[-1]

        # If transforms is not explicitly provided, infer automatically using
        # model args, kwargs.
        if self.transforms is None:
            self._set_transforms(*args, **kwargs)

        # transform samples back to constrained space
        for name, z in z_acc.items():
            if name in self.transforms:
                z_acc[name] = self.transforms[name].inv(z)
        self._samples = z_acc

        # terminate the sampler (shut down worker processes)
        self.sampler.terminate(True)

    def get_samples(self, num_samples=None, group_by_chain=False):
        """
        Get samples from the MCMC run, potentially resampling with replacement.

        For parameter details see: :meth:`select_samples <pyro.infer.mcmc.util.select_samples>`.
        """
        samples = self._samples
        return select_samples(samples, num_samples, group_by_chain)

    def diagnostics(self):
        """
        Gets some diagnostics statistics such as effective sample size, split
        Gelman-Rubin, or divergent transitions from the sampler.
        """
        diag = diagnostics(self._samples)
        for diag_name in self._diagnostics[0]:
            diag[diag_name] = {
                "chain {}".format(i): self._diagnostics[i][diag_name]
                for i in range(self.num_chains)
            }
        return diag

    def summary(self, prob=0.9):
        """
        Prints a summary table displaying diagnostics of samples obtained from
        posterior. The diagnostics displayed are mean, standard deviation, median,
        the 90% Credibility Interval, :func:`~pyro.ops.stats.effective_sample_size`,
        :func:`~pyro.ops.stats.split_gelman_rubin`.

        :param float prob: the probability mass of samples within the credibility interval.
        """
        print_summary(self._samples, prob=prob)
        if "divergences" in self._diagnostics[0]:
            print(
                "Number of divergences: {}".format(
                    sum(
                        [
                            len(self._diagnostics[i]["divergences"])
                            for i in range(self.num_chains)
                        ]
                    )
                )
            )


class StreamingMCMC(AbstractMCMC):
    """
    MCMC that computes required statistics in a streaming fashion. For this class no samples are retained
    but only aggregated statistics. This is useful for running memory expensive models where we care only
    about specific statistics (especially useful in a memory constrained environments like GPU).

    For available streaming ops please see :mod:`~pyro.ops.streaming`.
    """

    def __init__(
        self,
        kernel,
        num_samples,
        warmup_steps=None,
        initial_params=None,
        statistics=None,
        num_chains=1,
        hook_fn=None,
        disable_progbar=False,
        disable_validation=True,
        transforms=None,
        save_params=None,
    ):
        super().__init__(kernel, num_chains, transforms)
        self.warmup_steps = (
            num_samples if warmup_steps is None else warmup_steps
        )  # Stan
        self.num_samples = num_samples
        self.disable_validation = disable_validation
        self._samples = None
        self._args = None
        self._kwargs = None
        if statistics is None:
            statistics = StatsOfDict(default=CountMeanVarianceStats)
        self._statistics = statistics
        self._default_statistics = copy.deepcopy(statistics)
        if save_params is not None:
            kernel.save_params = save_params
        self._validate_kernel(initial_params)
        if num_chains > 1:
            if initial_params:
                self._validate_initial_params(initial_params)
        else:
            if initial_params:
                initial_params = {k: v.unsqueeze(0) for k, v in initial_params.items()}
        self._diagnostics = [None] * num_chains
        self.sampler = _UnarySampler(
            kernel,
            num_samples,
            self.warmup_steps,
            num_chains,
            disable_progbar,
            initial_params=initial_params,
            hook=hook_fn,
        )

    @poutine.block
    def run(self, *args, **kwargs):
        """
        Run StreamingMCMC to compute required `self._statistics`.
        """
        self._args, self._kwargs = args, kwargs
        num_samples = [0] * self.num_chains

        with optional(
            pyro.validation_enabled(not self.disable_validation),
            self.disable_validation is not None,
        ):
            args = [arg.detach() if torch.is_tensor(arg) else arg for arg in args]
            for x, chain_id in self.sampler.run(*args, **kwargs):
                if num_samples[chain_id] == 0:
                    # If transforms is not explicitly provided, infer automatically using
                    # model args, kwargs.
                    if self.transforms is None:
                        self._set_transforms(*args, **kwargs)
                    num_samples[chain_id] += 1
                    z_structure = x
                elif num_samples[chain_id] == self.num_samples + 1:
                    self._diagnostics[chain_id] = x
                else:
                    num_samples[chain_id] += 1
                    if self.num_chains > 1:
                        x_cloned = x.clone()
                        del x
                    else:
                        x_cloned = x

                    # unpack latent
                    pos = 0
                    z_acc = z_structure.copy()
                    for k in sorted(z_structure):
                        shape = z_structure[k]
                        next_pos = pos + shape.numel()
                        z_acc[k] = x_cloned[pos:next_pos].reshape(shape)
                        pos = next_pos

                    for name, z in z_acc.items():
                        if name in self.transforms:
                            z_acc[name] = self.transforms[name].inv(z)

                    self._statistics.update(
                        {
                            (chain_id, name): transformed_sample
                            for name, transformed_sample in z_acc.items()
                        }
                    )

        # terminate the sampler (shut down worker processes)
        self.sampler.terminate(True)

    def get_statistics(self, group_by_chain=True):
        """
        Returns a dict of statistics defined by those passed to the class constructor.

        :param bool group_by_chain: Whether statistics should be chain-wise or merged together.
        """
        if group_by_chain:
            return self._statistics.get()
        else:
            # merge all chains with respect to names
            merged_dict: Dict[str, StreamingStats] = {}
            for (_, name), stat in self._statistics.stats.items():
                if name in merged_dict:
                    merged_dict[name] = merged_dict[name].merge(stat)
                else:
                    merged_dict[name] = stat

        return {k: v.get() for k, v in merged_dict.items()}

    def diagnostics(self):
        """
        Gets diagnostics. Currently a split Gelman-Rubin is only supported and requires
        'mean' and 'variance' streaming statistics to be present.
        """
        statistics = self._statistics.get()
        diag = diagnostics_from_stats(statistics, self.num_samples, self.num_chains)
        for diag_name in self._diagnostics[0]:
            diag[diag_name] = {
                "chain {}".format(i): self._diagnostics[i][diag_name]
                for i in range(self.num_chains)
            }
        return diag



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import warnings

import torch

from pyro.distributions.torch import LKJCholesky, TransformedDistribution
from pyro.distributions.transforms.cholesky import CorrMatrixCholeskyTransform

from . import constraints


class LKJCorrCholesky(LKJCholesky):  # DEPRECATED
    def __init__(self, d, eta, validate_args=None):
        warnings.warn(
            "class LKJCorrCholesky(d, eta, validate_args=None) is deprecated "
            "in favor of LKJCholesky(dim, concentration, validate_args=None).",
            FutureWarning,
        )
        super().__init__(d, concentration=eta, validate_args=validate_args)


class LKJ(TransformedDistribution):
    r"""
    LKJ distribution for correlation matrices. The distribution is controlled by ``concentration``
    parameter :math:`\eta` to make the probability of the correlation matrix :math:`M` propotional
    to :math:`\det(M)^{\eta - 1}`. Because of that, when ``concentration == 1``, we have a
    uniform distribution over correlation matrices.

    When ``concentration > 1``, the distribution favors samples with large large determinent. This
    is useful when we know a priori that the underlying variables are not correlated.
    When ``concentration < 1``, the distribution favors samples with small determinent. This is
    useful when we know a priori that some underlying variables are correlated.

    :param int dimension: dimension of the matrices
    :param ndarray concentration: concentration/shape parameter of the
        distribution (often referred to as eta)

    **References**

    [1] `Generating random correlation matrices based on vines and extended onion method`,
    Daniel Lewandowski, Dorota Kurowicka, Harry Joe
    """

    arg_constraints = {"concentration": constraints.positive}
    support = constraints.corr_matrix

    def __init__(self, dim, concentration=1.0, validate_args=None):
        base_dist = LKJCholesky(dim, concentration)
        self.dim, self.concentration = base_dist.dim, base_dist.concentration
        super(LKJ, self).__init__(
            base_dist, CorrMatrixCholeskyTransform().inv, validate_args=validate_args
        )

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(LKJCholesky, _instance)
        return super(LKJCholesky, self).expand(batch_shape, _instance=new)

    @property
    def mean(self):
        return torch.eye(self.dim).expand(self.batch_shape + (self.dim, self.dim))



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

r"""
The pyro.infer.autoguide.initialization module contains initialization functions for
automatic guides.

The standard interface for initialization is a function that inputs a Pyro
trace ``site`` dict and returns an appropriately sized ``value`` to serve
as an initial constrained value for a guide estimate.
"""
import functools
from typing import Callable, Optional

import torch
from torch.distributions import transform_to

from pyro.distributions.torch import Independent
from pyro.distributions.torch_distribution import MaskedDistribution
from pyro.infer.util import is_validation_enabled
from pyro.poutine.messenger import Messenger
from pyro.util import torch_isnan

from .utils import helpful_support_errors

# TODO: move this file out of `autoguide` in a minor release


def _is_multivariate(d):
    while isinstance(d, (Independent, MaskedDistribution)):
        d = d.base_dist
    return any(size > 1 for size in d.event_shape)


def init_to_feasible(site=None):
    """
    Initialize to an arbitrary feasible point, ignoring distribution
    parameters.
    """
    if site is None:
        return init_to_feasible

    value = site["fn"].sample().detach()
    t = transform_to(site["fn"].support)
    value = t(torch.zeros_like(t.inv(value)))
    value._pyro_custom_init = False
    return value


def init_to_sample(site=None):
    """
    Initialize to a random sample from the prior.
    """
    if site is None:
        return init_to_sample

    value = site["fn"].sample().detach()
    value._pyro_custom_init = False
    return value


def init_to_median(
    site=None,
    num_samples=15,
    *,
    fallback: Optional[Callable] = init_to_feasible,
):
    """
    Initialize to the prior median; fallback to ``fallback`` (defaults to
    :func:`init_to_feasible`) if mean is undefined.

    :param callable fallback: Fallback init strategy, for sites not specified
        in ``values``.
    :raises ValueError: If ``fallback=None`` and no value for a site is given
        in ``values``.
    """
    if site is None:
        return functools.partial(
            init_to_median, num_samples=num_samples, fallback=fallback
        )

    # The median undefined for multivariate distributions.
    if _is_multivariate(site["fn"]):
        return init_to_feasible(site)
    try:
        # Try to compute empirical median.
        samples = site["fn"].sample(sample_shape=(num_samples,))
        value = samples.median(dim=0)[0]
        if torch_isnan(value):
            raise ValueError
        if hasattr(site["fn"], "_validate_sample"):
            site["fn"]._validate_sample(value)
        value._pyro_custom_init = False
        return value
    except (RuntimeError, ValueError):
        pass
    if fallback is not None:
        return fallback(site)
    raise ValueError(f"No init strategy specified for site {repr(site['name'])}")


def init_to_mean(
    site=None,
    *,
    fallback: Optional[Callable] = init_to_median,
):
    """
    Initialize to the prior mean; fallback to ``fallback`` (defaults to
    :func:`init_to_median`) if mean is undefined.

    :param callable fallback: Fallback init strategy, for sites not specified
        in ``values``.
    :raises ValueError: If ``fallback=None`` and no value for a site is given
        in ``values``.
    """
    if site is None:
        return functools.partial(init_to_mean, fallback=fallback)

    try:
        # Try .mean() method.
        value = site["fn"].mean.detach()
        if torch_isnan(value):
            raise ValueError
        if hasattr(site["fn"], "_validate_sample"):
            site["fn"]._validate_sample(value)
        value._pyro_custom_init = False
        return value
    except (NotImplementedError, ValueError):
        # This may happen for distributions with infinite variance, e.g. Cauchy.
        pass
    if fallback is not None:
        return fallback(site)
    raise ValueError(f"No init strategy specified for site {repr(site['name'])}")


def init_to_uniform(
    site: Optional[dict] = None,
    radius: float = 2.0,
):
    """
    Initialize to a random point in the area ``(-radius, radius)`` of
    unconstrained domain.

    :param float radius: specifies the range to draw an initial point in the
        unconstrained domain.
    """
    if site is None:
        return functools.partial(init_to_uniform, radius=radius)

    value = site["fn"].sample().detach()
    t = transform_to(site["fn"].support)
    value = t(torch.rand_like(t.inv(value)) * (2 * radius) - radius)
    value._pyro_custom_init = False
    return value


def init_to_value(
    site: Optional[dict] = None,
    values: dict = {},
    *,
    fallback: Optional[Callable] = init_to_uniform,
):
    """
    Initialize to the value specified in ``values``. Fallback to ``fallback``
    (defaults to :func:`init_to_uniform`) strategy for sites not appearing in
    ``values``.

    :param dict values: dictionary of initial values keyed by site name.
    :param callable fallback: Fallback init strategy, for sites not specified
        in ``values``.
    :raises ValueError: If ``fallback=None`` and no value for a site is given
        in ``values``.
    """
    if site is None:
        return functools.partial(init_to_value, values=values, fallback=fallback)

    if site["name"] in values:
        return values[site["name"]]
    if fallback is not None:
        return fallback(site)
    raise ValueError(f"No init strategy specified for site {repr(site['name'])}")


class _InitToGenerated:
    def __init__(self, generate):
        self.generate = generate
        self._init = None
        self._seen = set()

    def __call__(self, site):
        if self._init is None or site["name"] in self._seen:
            self._init = self.generate()
            self._seen = {site["name"]}
        return self._init(site)


def init_to_generated(site=None, generate=lambda: init_to_uniform):
    """
    Initialize to another initialization strategy returned by the callback
    ``generate`` which is called once per model execution.

    This is like :func:`init_to_value` but can produce different (e.g. random)
    values once per model execution. For example to generate values and return
    ``init_to_value`` you could define::

        def generate():
            values = {"x": torch.randn(100), "y": torch.rand(5)}
            return init_to_value(values=values)

        my_init_fn = init_to_generated(generate=generate)

    :param callable generate: A callable returning another initialization
        function, e.g. returning an ``init_to_value(values={...})`` populated
        with a dictionary of random samples.
    """
    init = _InitToGenerated(generate)
    return init if site is None else init(site)


class InitMessenger(Messenger):
    """
    Initializes a site by replacing ``.sample()`` calls with values
    drawn from an initialization strategy. This is mainly for internal use by
    autoguide classes.

    :param callable init_fn: An initialization function.
    """

    def __init__(self, init_fn):
        self.init_fn = init_fn
        super().__init__()

    def _pyro_sample(self, msg):
        if msg["value"] is not None or type(msg["fn"]).__name__ == "_Subsample":
            return
        with torch.no_grad(), helpful_support_errors(msg):
            value = self.init_fn(msg)
        if is_validation_enabled() and msg["value"] is not None:
            if not isinstance(value, type(msg["value"])):
                raise ValueError(
                    "{} provided invalid type for site {}:\nexpected {}\nactual {}".format(
                        self.init_fn, msg["name"], type(msg["value"]), type(value)
                    )
                )
            if value.shape != msg["value"].shape:
                raise ValueError(
                    "{} provided invalid shape for site {}:\nexpected {}\nactual {}".format(
                        self.init_fn, msg["name"], msg["value"].shape, value.shape
                    )
                )
        msg["value"] = value

    def _pyro_get_init_messengers(self, msg):
        if msg["value"] is None:
            msg["value"] = []
        msg["value"].append(self)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import math

import torch
from torch.distributions import constraints
from torch.distributions.utils import broadcast_all

from .torch_distribution import TorchDistribution


class SoftLaplace(TorchDistribution):
    """
    Smooth distribution with Laplace-like tail behavior.

    This distribution corresponds to the log-convex density::

        z = (value - loc) / scale
        log_prob = log(2 / pi) - log(scale) - logaddexp(z, -z)

    Like the Laplace density, this density has the heaviest possible tails
    (asymptotically) while still being log-convex. Unlike the Laplace
    distribution, this distribution is infinitely differentiable everywhere,
    and is thus suitable for constructing Laplace approximations.

    :param loc: Location parameter.
    :param scale: Scale parameter.
    """

    arg_constraints = {"loc": constraints.real, "scale": constraints.positive}
    support = constraints.real
    has_rsample = True

    def __init__(self, loc, scale, *, validate_args=None):
        self.loc, self.scale = broadcast_all(loc, scale)
        super().__init__(self.loc.shape, validate_args=validate_args)

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(SoftLaplace, _instance)
        batch_shape = torch.Size(batch_shape)
        new.loc = self.loc.expand(batch_shape)
        new.scale = self.scale.expand(batch_shape)
        super(SoftLaplace, new).__init__(batch_shape, validate_args=False)
        new._validate_args = self._validate_args
        return new

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)
        z = (value - self.loc) / self.scale
        return math.log(2 / math.pi) - self.scale.log() - torch.logaddexp(z, -z)

    def rsample(self, sample_shape=torch.Size()):
        shape = self._extended_shape(sample_shape)
        u = self.loc.new_empty(shape).uniform_()
        return self.icdf(u)

    def cdf(self, value):
        if self._validate_args:
            self._validate_sample(value)
        z = (value - self.loc) / self.scale
        return z.exp().atan().mul(2 / math.pi)

    def icdf(self, value):
        return value.mul(math.pi / 2).tan().log().mul(self.scale).add(self.loc)

    @property
    def mean(self):
        return self.loc

    @property
    def variance(self):
        return (math.pi / 2 * self.scale) ** 2
